"""Unit tests for the DraftFileDiffResource."""

from __future__ import annotations

import os

from django.core.files.uploadedfile import SimpleUploadedFile
from djblets.webapi.errors import (INVALID_FORM_DATA,
                                   PERMISSION_DENIED)
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard import scmtools
from reviewboard.attachments.models import FileAttachment
from reviewboard.diffviewer.models import DiffSet, FileDiff
from reviewboard.scmtools.core import PRE_CREATION
from reviewboard.webapi.errors import DIFF_TOO_BIG
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (diff_item_mimetype,
                                                filediff_item_mimetype,
                                                filediff_list_mimetype)
from reviewboard.webapi.tests.mixins import BasicTestsMetaclass
from reviewboard.webapi.tests.mixins_extra_data import ExtraDataItemMixin
from reviewboard.webapi.tests.urls import (get_diff_list_url,
                                           get_draft_filediff_item_url,
                                           get_draft_filediff_list_url)


class ResourceListTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the DraftFileDiffResource list APIs."""
    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-requests/<id>/draft/diffs/<revision>/files/'
    resource = resources.draft_filediff

    def compare_item(self, item_rsp, filediff):
        self.assertEqual(item_rsp['id'], filediff.pk)
        self.assertEqual(item_rsp['source_file'], filediff.source_file)

        filtered_extra_data = {
            key: value
            for key, value in filediff.extra_data.items()
            if not key.startswith('_')
        }

        self.assertEqual(item_rsp['extra_data'], filtered_extra_data)

    def setup_http_not_allowed_item_test(self, user):
        review_request = self.create_review_request(
            create_repository=True,
            submitter=user)
        diffset = self.create_diffset(review_request, draft=True)

        return get_draft_filediff_list_url(diffset, review_request)

    def setup_http_not_allowed_list_test(self, user):
        review_request = self.create_review_request(
            create_repository=True,
            submitter=user)
        diffset = self.create_diffset(review_request, draft=True)

        return get_draft_filediff_list_url(diffset, review_request)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user)
        diffset = self.create_diffset(review_request, draft=True)

        if populate_items:
            items = [self.create_filediff(diffset)]
        else:
            items = []

        return (get_draft_filediff_list_url(diffset, review_request,
                                            local_site_name),
                filediff_list_mimetype,
                items)

    def test_get_not_owner(self):
        """Testing the
        GET review-requests/<id>/draft/diffs/<revision>/files/ API
        without owner with Permission Denied error
        """
        review_request = self.create_review_request(create_repository=True)
        self.assertNotEqual(review_request.submitter, self.user)
        diffset = self.create_diffset(review_request, draft=True)

        self.api_get(
            get_draft_filediff_list_url(diffset, review_request),
            expected_status=403)

    @webapi_test_template
    def test_commit_filter(self):
        """Testing the GET <URL>?commit-id= API filters draft FileDiffs to the
        requested commit
        """
        repository = self.create_repository()
        review_request = self.create_review_request(repository=repository,
                                                    submitter=self.user)
        diffset = self.create_diffset(review_request=review_request,
                                      draft=True,
                                      repository=repository)
        commit = self.create_diffcommit(diffset=diffset,
                                        repository=repository)

        rsp = self.api_get(
            '%s?commit-id=%s'
            % (get_draft_filediff_list_url(diffset, review_request),
               commit.commit_id),
            expected_status=200,
            expected_mimetype=filediff_list_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('files', rsp)
        self.assertEqual(rsp['total_results'], 1)

        item_rsp = rsp['files'][0]
        filediff = FileDiff.objects.get(pk=item_rsp['id'])
        self.compare_item(item_rsp, filediff)

    @webapi_test_template
    def test_commit_filter_no_results(self):
        """Testing the GET <URL>?commit-id= API with no results"""
        repository = self.create_repository()
        review_request = self.create_review_request(repository=repository,
                                                    submitter=self.user)
        diffset = self.create_diffset(review_request=review_request,
                                      draft=True,
                                      repository=repository)
        commit = self.create_diffcommit(diffset=diffset,
                                        repository=repository)

        rsp = self.api_get(
            '%s?commit-id=%s'
            % (get_draft_filediff_list_url(diffset, review_request),
               commit.parent_id),
            expected_status=200,
            expected_mimetype=filediff_list_mimetype)

        self.assertIn('stat', rsp)
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('files', rsp)
        self.assertEqual(rsp['files'], [])
        self.assertEqual(rsp['total_results'], 0)


class ResourceItemTests(ExtraDataItemMixin, BaseWebAPITestCase,
                        metaclass=BasicTestsMetaclass):
    """Testing the DraftFileDiffResource item APIs."""
    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-requests/<id>/draft/diffs/<revision>/files/<id>/'
    resource = resources.draft_filediff
    test_http_methods = ('DELETE', 'GET', 'PUT')

    def setup_http_not_allowed_item_test(self, user):
        review_request = self.create_review_request(
            create_repository=True,
            submitter=user)
        diffset = self.create_diffset(review_request, draft=True)
        filediff = self.create_filediff(diffset)

        return get_draft_filediff_item_url(filediff, review_request)

    def compare_item(self, item_rsp, filediff):
        self.assertEqual(item_rsp['id'], filediff.pk)
        self.assertEqual(item_rsp['source_file'], filediff.source_file)
        self.assertEqual(item_rsp['extra_data'], filediff.extra_data)
        self.assertEqual(item_rsp['orig_sha256'], filediff.orig_sha256)
        self.assertEqual(item_rsp['patched_sha256'], filediff.patched_sha256)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user)
        diffset = self.create_diffset(review_request, draft=True)
        filediff = self.create_filediff(diffset)

        return (get_draft_filediff_item_url(filediff, review_request,
                                            local_site_name),
                filediff_item_mimetype,
                filediff)

    def test_get_not_owner(self):
        """Testing the
        GET review-requests/<id>/draft/diffs/<revision>/files/<id>/ API
        without owner with Permission Denied error
        """
        review_request = self.create_review_request(create_repository=True)
        self.assertNotEqual(review_request.submitter, self.user)
        diffset = self.create_diffset(review_request, draft=True)
        filediff = self.create_filediff(diffset)

        self.api_get(
            get_draft_filediff_item_url(filediff, review_request),
            expected_status=403)

    #
    # HTTP PUT tests
    #

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        review_request = self.create_review_request(
            submitter=user,
            with_local_site=with_local_site,
            create_repository=True)
        diffset = self.create_diffset(review_request, draft=True)
        filediff = self.create_filediff(diffset)

        return (get_draft_filediff_item_url(filediff, review_request,
                                            local_site_name),
                filediff_item_mimetype,
                {},
                filediff,
                [])

    def check_put_result(self, user, item_rsp, filediff):
        filediff = FileDiff.objects.get(pk=filediff.pk)
        self.compare_item(item_rsp, filediff)

    def test_put_with_new_file_and_dest_attachment_file(self):
        """Testing the PUT review-requests/<id>/diffs/<id>/files/<id>/ API
        with new file and dest_attachment_file
        """
        review_request = self.create_review_request(create_repository=True,
                                                    submitter=self.user)

        diff_filename = os.path.join(os.path.dirname(scmtools.__file__),
                                     'testdata', 'git_binary_image_new.diff')

        with open(diff_filename, 'rb') as f:
            rsp = self.api_post(
                get_diff_list_url(review_request),
                {
                    'path': f,
                    'base_commit_id': '1234',
                },
                expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        diffset = DiffSet.objects.get(pk=rsp['diff']['id'])
        filediffs = diffset.files.all()

        self.assertEqual(len(filediffs), 1)
        filediff = filediffs[0]
        self.assertEqual(filediff.source_file, 'logo.png')

        with open(self.get_sample_image_filename(), 'rb') as f:
            rsp = self.api_put(
                get_draft_filediff_item_url(filediff, review_request) +
                '?expand=dest_attachment',
                {
                    'dest_attachment_file': f,
                },
                expected_mimetype=filediff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('dest_attachment', rsp['file'])

        attachment = FileAttachment.objects.get(
            pk=rsp['file']['dest_attachment']['id'])

        self.assertTrue(attachment.is_from_diff)
        self.assertEqual(attachment.orig_filename, 'logo.png')
        self.assertEqual(attachment.added_in_filediff, filediff)
        self.assertEqual(attachment.repo_path, None)
        self.assertEqual(attachment.repo_revision, None)
        self.assertEqual(attachment.repository, None)

    def test_put_with_modified_file_and_dest_attachment_file(self):
        """Testing the PUT review-requests/<id>/diffs/<id>/files/<id>/ API
        with modified file and dest_attachment_file
        """
        review_request = self.create_review_request(create_repository=True,
                                                    submitter=self.user)

        diff = SimpleUploadedFile('git_binary_image_modified.diff',
                                  self.DEFAULT_GIT_BINARY_IMAGE_DIFF,
                                  content_type='text/x-patch')
        rsp = self.api_post(
            get_diff_list_url(review_request),
            {
                'path': diff,
                'base_commit_id': '1234',
            },
            expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        diffset = DiffSet.objects.get(pk=rsp['diff']['id'])
        filediffs = diffset.files.all()

        self.assertEqual(len(filediffs), 1)
        filediff = filediffs[0]
        self.assertEqual(filediff.source_file, 'logo.png')

        with open(self.get_sample_image_filename(), 'rb') as f:
            rsp = self.api_put(
                get_draft_filediff_item_url(filediff, review_request) +
                '?expand=dest_attachment',
                {
                    'dest_attachment_file': f,
                },
                expected_mimetype=filediff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('dest_attachment', rsp['file'])

        attachment = FileAttachment.objects.get(
            pk=rsp['file']['dest_attachment']['id'])

        self.assertTrue(attachment.is_from_diff)
        self.assertEqual(attachment.orig_filename, 'logo.png')
        self.assertEqual(attachment.added_in_filediff, filediff)

    @webapi_test_template
    def test_put_attachment_file_too_big(self) -> None:
        """Testing the PUT <URL> API with a file that exceeds configured size
        limits
        """
        review_request = self.create_review_request(create_repository=True,
                                                    submitter=self.user)

        diff = SimpleUploadedFile('git_binary_image_modified.diff',
                                  self.DEFAULT_GIT_BINARY_IMAGE_DIFF,
                                  content_type='text/x-patch')
        rsp = self.api_post(
            get_diff_list_url(review_request),
            {
                'path': diff,
                'base_commit_id': '1234',
            },
            expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        diffset = DiffSet.objects.get(pk=rsp['diff']['id'])
        filediffs = diffset.files.all()

        self.assertEqual(len(filediffs), 1)
        filediff = filediffs[0]
        self.assertEqual(filediff.source_file, 'logo.png')

        with self.siteconfig_settings({'diffviewer_max_binary_size': 2},
                                      reload_settings=False):
            with open(self.get_sample_image_filename(), 'rb') as f:
                rsp = self.api_put(
                    get_draft_filediff_item_url(filediff, review_request) +
                    '?expand=dest_attachment',
                    {
                        'dest_attachment_file': f,
                    },
                    expected_status=400)

        assert rsp is not None
        self.assertEqual(rsp, {
            'err': {
                'code': DIFF_TOO_BIG.code,
                'msg': 'The specified diff file is too large.',
                'type': 'diff-too-large',
            },
            'max_size': 2,
            'reason': 'The given file is too large.',
            'stat': 'fail',
        })

    def test_put_second_dest_attachment_file_disallowed(self):
        """Testing the PUT review-requests/<id>/diffs/<id>/files/<id>/ API
        disallows setting dest_attachment_file twice
        """
        review_request = self.create_review_request(create_repository=True,
                                                    submitter=self.user)

        diff = SimpleUploadedFile('git_binary_image_modified.diff',
                                  self.DEFAULT_GIT_BINARY_IMAGE_DIFF,
                                  content_type='text/x-patch')

        rsp = self.api_post(
            get_diff_list_url(review_request),
            {
                'path': diff,
                'base_commit_id': '1234',
            },
            expected_mimetype=diff_item_mimetype)

        diffset = DiffSet.objects.get(pk=rsp['diff']['id'])
        filediff = diffset.files.all()[0]

        url = get_draft_filediff_item_url(filediff, review_request)
        trophy_filename = self.get_sample_image_filename()

        with open(trophy_filename, 'rb') as f:
            self.api_put(
                url,
                {
                    'dest_attachment_file': f,
                },
                expected_mimetype=filediff_item_mimetype)

        with open(trophy_filename, 'rb') as f:
            rsp = self.api_put(
                url,
                {
                    'dest_attachment_file': f,
                },
                expected_status=400)

            self.assertEqual(rsp['stat'], 'fail')
            self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
            self.assertIn('fields', rsp)
            self.assertIn('dest_attachment_file', rsp['fields'])

    @webapi_test_template
    def test_get_with_diff_data_and_inaccessible(self) -> None:
        """Testing the GET <URL> API with diff data result and inaccessible
        FileDiff
        """
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository)

        self.assertNotEqual(self.user, review_request.owner)
        self.assertFalse(review_request.is_accessible_by(self.user))

        diffset = self.create_diffset(review_request, draft=True)
        filediff = self.create_filediff(
            diffset,
            source_file='newfile.py',
            source_revision=PRE_CREATION,
            dest_file='newfile.py',
            dest_detail='20e43bb7c2d9f3a31768404ac71121804d806f7c',
            diff=(
                b"diff --git a/newfile.py b/newfile.py\n"
                b"new file mode 100644\n"
                b"index 0000000000000000000000000000000000000000.."
                b"8eaa5c1eacb55c43f5e00ed9dcd0c8da901f0c85\n"
                b"--- /dev/null\n"
                b"+++ b/newfile.py\n"
                b"@@ -0,0 +1 @@\n"
                b"+print('hello, world!')\n"
            ))

        rsp = self.api_get(
            get_draft_filediff_item_url(filediff, review_request),
            HTTP_ACCEPT='application/vnd.reviewboard.org.diff.data+json',
            expected_status=403)

        self.assertEqual(
            rsp,
            {
                'err': {
                    'code': PERMISSION_DENIED.code,
                    'msg': PERMISSION_DENIED.msg,
                    'type': 'resource-permission-denied',
                },
                'stat': 'fail',
            })

    @webapi_test_template
    def test_get_with_patch_and_inaccessible(self) -> None:
        """Testing the GET <URL> API with patch result and inaccessible
        FileDiff
        """
        repository = self.create_repository(tool_name='Git')
        review_request = self.create_review_request(repository=repository)

        self.assertNotEqual(self.user, review_request.owner)
        self.assertFalse(review_request.is_accessible_by(self.user))

        diffset = self.create_diffset(review_request, draft=True)
        filediff = self.create_filediff(
            diffset,
            source_file='newfile.py',
            source_revision=PRE_CREATION,
            dest_file='newfile.py',
            dest_detail='20e43bb7c2d9f3a31768404ac71121804d806f7c',
            diff=(
                b"diff --git a/newfile.py b/newfile.py\n"
                b"new file mode 100644\n"
                b"index 0000000000000000000000000000000000000000.."
                b"8eaa5c1eacb55c43f5e00ed9dcd0c8da901f0c85\n"
                b"--- /dev/null\n"
                b"+++ b/newfile.py\n"
                b"@@ -0,0 +1 @@\n"
                b"+print('hello, world!')\n"
            ))

        rsp = self.api_get(
            get_draft_filediff_item_url(filediff, review_request),
            HTTP_ACCEPT='text/x-patch',
            expected_status=403)

        self.assertEqual(
            rsp,
            {
                'err': {
                    'code': PERMISSION_DENIED.code,
                    'msg': PERMISSION_DENIED.msg,
                    'type': 'resource-permission-denied',
                },
                'stat': 'fail',
            })



"""API resource for user session management."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING

from django.contrib.auth import logout
from django.contrib.auth.models import User
from django.utils.translation import gettext as _
from djblets.util.json_utils import json_merge_patch
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_request_fields)
from djblets.webapi.errors import INVALID_FORM_DATA
from djblets.webapi.resources.registry import get_resource_for_object

from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.decorators import (webapi_check_login_required,
                                           webapi_check_local_site)

if TYPE_CHECKING:
    from django.http import HttpRequest

    from djblets.webapi.resources.base import WebAPIResourceHandlerResult
    from djblets.webapi.responses import WebAPIResponsePayload


class SessionResource(WebAPIResource):
    """Information on the active user's session.

    This includes information on the user currently logged in through the
    calling client, if any. Currently, the resource links to that user's
    own resource, making it easy to figure out the user's information and
    any useful related resources.
    """

    name = 'session'
    singleton = True
    allowed_methods = ('GET', 'PUT', 'DELETE')

    #: Settings paths that can be modified in a PUT.
    #:
    #: Version Added:
    #:     8.0
    _MUTABLE_PROFILE_SETTING_PATHS: set[tuple[str, ...]] = {
        ('confirm_ship_it',),
        ('quick_access_action_ids',),
    }

    def serialize_object(
        self,
        obj: None,
        request: (HttpRequest | None) = None,
        *args,
        **kwargs,
    ) -> WebAPIResponsePayload:
        """Serialize the session resource.

        This specially serializes some state and links for the session
        based on the login state.

        Args:
            obj (object):
                The object to serialize.

                This will always be ``None`` for this resource.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments passed to the view.

            **kwargs (dict):
                Keyword arguments passed to the view.

        Returns:
            djblets.webapi.responses.WebAPIResponsePayload:
            The serialized payload.
        """
        assert request is not None

        expanded_resources = (
            request.GET.get('expand', '') or
            request.POST.get('expand', '')
        ).split(',')

        user = request.user
        authenticated = user.is_authenticated

        data: WebAPIResponsePayload = {
            'authenticated': authenticated,
            'links': self.get_links(request=request, *args, **kwargs),
        }

        if authenticated and 'user' in expanded_resources:
            data['user'] = user
            del data['links']['user']

        return data

    @webapi_check_local_site
    @webapi_check_login_required
    def get(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> WebAPIResourceHandlerResult:
        """Returns information on the client's session.

        This currently just contains information on the currently logged-in
        user (if any).
        """
        return 200, {
            self.name: self.serialize_object(obj=None,
                                             request=request,
                                             *args, **kwargs),
        }

    @webapi_check_local_site
    @webapi_login_required
    def delete(self, request, *args, **kwargs):
        """Clears the user's client session and the session cookie.

        This is equivalent to logging out a user. The existing session cookie
        will be invalidated and will no longer be accepted.

        This will return a :http:`204`.
        """
        logout(request)

        return 204, {}

    @webapi_login_required
    @webapi_check_local_site
    @webapi_request_fields(
        optional={
            'settings:json': {
                'type': str,
                'description': (
                    'A JSON Merge Patch of settings change to make. These '
                    'will be persisted across sessions. This is considered '
                    'internal API and is subject to change.'
                ),
            },
        },
    )
    def update(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> WebAPIResourceHandlerResult:
        """Update information about the session.

        This is only used internally to manage Review Board session state.
        This operation is not considered public API.

        Version Added:
            8.0
        """
        settings_json = request.POST.get('settings:json')

        if settings_json:
            try:
                patch = json.loads(settings_json)
            except ValueError as e:
                return INVALID_FORM_DATA, {
                    'fields': {
                        'settings': [
                            # Use %-based formatting to share translations
                            # with ImportExtraDataError.error_payload.
                            _('Could not parse JSON data: %s') % e,
                        ],
                    },
                }

            user = request.user
            assert isinstance(user, User)

            profile = user.get_profile(create_if_missing=True)

            MUTABLE_SETTING_PATHS = self._MUTABLE_PROFILE_SETTING_PATHS

            new_settings = json_merge_patch(
                profile.settings,
                patch,
                can_write_key_func=lambda *, path, **kwargs:
                    path in MUTABLE_SETTING_PATHS)

            # Save extra_data only if it remains a dictionary, so callers
            # can't replace the entire contents.
            if not isinstance(new_settings, dict):
                return INVALID_FORM_DATA, {
                    'fields': {
                        'settings': [
                            _('settings:json cannot replace the settings '
                              'with a non-dictionary type'),
                        ],
                    },
                }

            profile.settings.clear()
            profile.settings.update(new_settings)
            profile.save(update_fields=('settings',))

        return 200, {
            self.name: self.serialize_object(obj=None,
                                             request=request,
                                             *args, **kwargs),
        }

    def get_related_links(self, obj=None, request=None, *args, **kwargs):
        links = {}

        if request and request.user.is_authenticated:
            user_resource = get_resource_for_object(request.user)
            href = user_resource.get_href(request.user, request,
                                          *args, **kwargs)

            # Since there's no object, DELETE won't be populated automatically.
            clean_href = request.build_absolute_uri()
            i = clean_href.find('?')

            if i != -1:
                clean_href = clean_href[:i]

            links['delete'] = {
                'method': 'DELETE',
                'href': clean_href,
            }

            links['user'] = {
                'method': 'GET',
                'href': href,
                'title': str(request.user),
                'resource': user_resource,
                'list-resource': False,
            }

        return links


session_resource = SessionResource()



"""Base hosting service configuration forms.

Version Added:
    6.0:
    This replaces the old :py:mod:`reviewboard.hostingsvcs.forms` module.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django import forms
from django.utils.translation import gettext, gettext_lazy as _
from typelets.runtime import raise_invalid_type

from reviewboard.certs.errors import CertificateVerificationError
from reviewboard.certs.manager import cert_manager
from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            TwoFactorAuthCodeRequiredError)
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.scmtools.errors import (
    UnverifiedCertificateError as LegacyUnverifiedCertificateError,
)
from reviewboard.scmtools.forms import (BaseRepositoryAuthSubForm,
                                        BaseRepositoryInfoSubForm)

if TYPE_CHECKING:
    from typelets.funcs import KwargsDict
    from typelets.json import JSONDict

    from reviewboard.hostingsvcs.base import BaseHostingService
    from reviewboard.hostingsvcs.base.hosting_service import \
        HostingServiceCredentials
    from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


class _HostingServiceSubFormMixin:
    """Mixin for hosting service subforms.

    This is used internally by :py:class:`HostingServiceForm` and
    :py:class:`HostingServiceAuthForm` to check and set initial state
    relating to the hosting service.

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.forms` to
          :py:mod:`reviewboard.hostingsvcs.base.forms`.

    Version Added:
        3.0.16
    """

    ######################
    # Instance variables #
    ######################

    #: The hosting service class that owns this form.
    #:
    #: Type:
    #:     type
    hosting_service_cls: type[BaseHostingService]

    def __init__(
        self,
        *args,
        hosting_service_cls: type[BaseHostingService],
        **kwargs,
    ) -> None:
        """Initialize the authentication form.

        Args:
            *args (tuple):
                Additional positional arguments for the parent form.

            hosting_service_cls (type):
                The hosting service class that that works with this form.

            **kwargs (dict):
                Additional keyword arguments for the parent form.

        Raises:
            ValueError:
                One or more of the parameters are missing or not valid for the
                provided hosting account. Details are given in the error
                message.
        """
        if not hosting_service_cls:
            raise_invalid_type(hosting_service_cls,
                               'hosting_service_cls cannot be None.')

        self.hosting_service_cls = hosting_service_cls

        super().__init__(*args, **kwargs)


class BaseHostingServiceAuthForm(_HostingServiceSubFormMixin,
                                 BaseRepositoryAuthSubForm):
    """Base form for handling authentication information for a hosting account.

    This takes care of collecting additional details needed for authenticating
    an account, including that information with the account credentials (if
    needed by the hosting service).

    By default, this will retain the existing username, password, and
    two-factor auth fields. Those can be replaced, but the field names should
    remain the same.

    Unlike :py:class:`HostingServiceForm`, field names on this class do not
    need to include a service-specific prefix, as they will not conflict with
    other forms. The field names will be used for the data storage. How a
    subclass chooses to name these fields is up to them.

    Subclasses can define a ``Meta`` class on the form containing
    ``help_texts`` and ``labels`` attributes, mapping field names to custom
    help text or labels. This is useful for providing more specific
    instructions for setting authentication data for a given service without
    having to override the built-in fields. For example:

    .. code-block:: python

        from django.utils.translation import gettext_lazy as _
        from reviewboard.hostingsvcs.base.forms import \
            BaseHostingServiceAuthForm


        class MyAuthForm(HostingServiceAuthForm):
            class Meta:
                labels = {
                    'hosting_account_username': 'API Access ID',
                    'hosting_account_password': 'API Secret Key',
                }

                help_texts = {
                    'hosting_account_username': _(
                        'Access ID used for the API. This can be found in '
                        'your FooService account settings.'
                    ),
                    'hosting_account_password': _(
                        'Secret key used for the API. This can be found in '
                        'your FooService account settings.'
                    ),
                }

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.forms` to
          :py:mod:`reviewboard.hostingsvcs.base.forms` and renamed from
          ``HostingServiceAuthForm`` to ``BaseHostingServiceAuthForm``.
    """

    hosting_url = forms.CharField(
        label=_('Service URL'),
        required=True,
        widget=forms.TextInput(attrs={'size': 30}))

    hosting_account_username = forms.CharField(
        label=_('Account username'),
        required=True,
        widget=forms.TextInput(attrs={
            'size': 30,
            'autocomplete': 'off',
        }))

    hosting_account_password = forms.CharField(
        label=_('Account password'),
        required=True,
        widget=forms.PasswordInput(
            attrs={
                'size': 30,
                'autocomplete': 'off',
            },
            render_value=True))

    hosting_account_two_factor_auth_code = forms.CharField(
        label=_('Two-factor auth code'),
        required=False,
        widget=forms.TextInput(attrs={
            'size': 30,
            'autocomplete': 'off',
            'data-required-for-2fa': 'true',
        }))

    ######################
    # Instance variables #
    ######################

    #: The hosting service account being configured.
    #:
    #: If ``None``, a new one will be created when saving the form.
    #:
    #: Type:
    #:     reviewboard.hostingsvcs.models.HostingServiceAccount
    hosting_account: HostingServiceAccount | None

    def __init__(
        self,
        *args,
        hosting_account: (HostingServiceAccount | None) = None,
        **kwargs,
    ) -> None:
        """Initialize the authentication form.

        Args:
            *args (tuple):
                Additional positional arguments for the parent form.

            hosting_account (reviewboard.hostingsvcs.models.
                             HostingServiceAccount, optional):
                The hosting service account being updated, if any.

                If ``None``, a new one will be created.

            **kwargs (dict):
                Additional keyword arguments for the parent form.

        Raises:
            ValueError:
                One or more of the parameters are missing or not valid for the
                provided hosting account. Details are given in the error
                message.
        """
        self.hosting_account = hosting_account

        super().__init__(*args, **kwargs)

        hosting_service_cls = self.hosting_service_cls
        hosting_service_id = hosting_service_cls.hosting_service_id

        # Make sure that the hosting account, if provided, is compatible with
        # the arguments provided.
        if (hosting_account and
            (hosting_account.local_site != self.local_site or
             hosting_account.service_name != hosting_service_id)):
            raise ValueError(
                gettext('This account is not compatible with this '
                        'hosting service configuration.'))

        # If the hosting service is not self-hosted, we don't want to include
        # the hosting_url form.
        if not hosting_service_cls.self_hosted:
            del self.fields['hosting_url']

        # If it doesn't support two-factor auth, get rid of that field.
        if not hosting_service_cls.supports_two_factor_auth:
            del self.fields['hosting_account_two_factor_auth_code']

    def get_initial_data(self) -> JSONDict:
        """Return initial data for the form, based on the hosting account.

        This will return initial data for the fields, generally pulled from
        the hosting account. This will be used when relinking a hosting
        account that's no longer authorized.

        Generally, sensitive information, like passwords, should not be
        provided.

        By default, the :py:attr:`username` and :py:attr:`hosting_url` fields
        will have data provided. Subclasses can override this to present more
        initial data.

        This is only called if the form was provided a hosting account during
        construction.

        Returns:
            dict:
            Initial data for the form.
        """
        initial: JSONDict = {}

        if self.hosting_account:
            initial['username'] = self.hosting_account.username

            if self.hosting_service_cls.self_hosted:
                initial['hosting_url'] = self.hosting_account.hosting_url

        return initial

    def get_credentials(self) -> HostingServiceCredentials:
        """Return credentials from the form.

        This should return the data that will be stored along with the
        :py:class:`~reviewboard.hostingsvcs.models.HostingServiceAccount`.

        The ``username``, ``password``, and ``two_factor_auth_code`` values
        are treated specially during the creation and authentication of the
        account, and should be provided for most standard hosting services.

        All values will be provided to
        :py:attr:`BaseHostingService.authenticate <reviewboard.hostingsvcs.
        base.hosting_service.HostingService.authenticate>`, which will be
        responsible for making use of these values and storing them on the
        account.

        Subclasses should call the parent method and use their results as
        a base, if they reuse any of the built-in fields.

        Returns:
            dict:
            A dictionary of credentials used to authenticate the account and
            talk to the API.
        """
        credentials: HostingServiceCredentials = {
            'username': self.cleaned_data['hosting_account_username'],
            'password': self.cleaned_data['hosting_account_password'],
        }

        two_factor_auth_code = \
            self.cleaned_data.get('hosting_account_two_factor_auth_code')

        if two_factor_auth_code:
            credentials['two_factor_auth_code'] = two_factor_auth_code

        return credentials

    def save(
        self,
        allow_authorize: bool = True,
        force_authorize: bool = False,
        extra_authorize_kwargs: (KwargsDict | None) = None,
        trust_host: bool = False,
        save: bool = True,
    ) -> HostingServiceAccount:
        """Save the hosting account and authorize against the service.

        This will create or update a hosting account, based on the information
        provided in the form and to this method.

        :py:meth:`is_valid` must be called prior to saving.

        Args:
            allow_authorize (bool, optional):
                If ``True`` (the default), the account will be authorized
                against the hosting service. If ``False``, only the database
                entry for the account will be affected.

            force_authorize (bool, optional):
                Force the account to be re-authorized, if already authorized.

            extra_authorize_kwargs (dict, optional):
                Additional keyword arguments to provide for the
                :py:meth:`HostingService.authorize()
                <reviewboard.hostingsvcs.models.HostingService.authorize>`
                call.

            trust_host (bool, optional):
                Whether to trust the given host, even if the linked certificate
                is invalid or self-signed.

            save (bool, optional):
                Whether or not the created account should be saved.

                This is intended to be used by subclasses who want to add
                additional data to the resulting hosting account before saving.

                If this is ``False``, the caller must ensure the resulting
                hosting account is saved.

        Returns:
            reviewboard.hostingsvcs.models.HostingServiceAccount:
            The updated or created hosting service account.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                Information needed to authorize was missing, or authorization
                failed.

            reviewboard.hostingsvcs.errors.TwoFactorAuthCodeRequiredError:
                A two-factor authentication code is required to authorize the
                account. A code will need to be provided to the form.
        """
        if extra_authorize_kwargs is None:
            extra_authorize_kwargs = {}

        credentials = self.get_credentials()

        # Grab the username from the credentials, sanity-checking that it's
        # been provided as part of the get_credentials() result.
        try:
            username = credentials['username']
        except KeyError:
            logger.exception('%s.get_credentials() must return a "username" '
                             'key.',
                             self.__class__.__name__)

            raise AuthorizationError(
                gettext('Hosting service implementation error: '
                        '%s.get_credentials() must return a "username" key.')
                % self.__class__.__name__)

        hosting_account = self.hosting_account
        hosting_service_id = self.hosting_service_cls.hosting_service_id
        hosting_url: (str | None) = self.cleaned_data.get('hosting_url')

        assert hosting_service_id

        if not self.hosting_service_cls.self_hosted:
            assert hosting_url is None

        if hosting_account:
            # Update the username and hosting URL, if they've changed.
            hosting_account.username = username
            hosting_account.hosting_url = hosting_url
        else:
            # Fetch an existing hosting account based on the credentials and
            # parameters, if there is one. If not, we're going to create one,
            # but we won't save it until we've authorized.
            hosting_account_attrs: KwargsDict = {
                'service_name': hosting_service_id,
                'username': username,
                'hosting_url': hosting_url,
                'local_site': self.local_site,
            }

            try:
                hosting_account = \
                    HostingServiceAccount.objects.get(**hosting_account_attrs)
            except HostingServiceAccount.DoesNotExist:
                # Create a new one, but don't save it yet.
                hosting_account = \
                    HostingServiceAccount(**hosting_account_attrs)

        assert hosting_account is not None

        if (allow_authorize and
            self.hosting_service_cls.needs_authorization and
            (not hosting_account.is_authorized or force_authorize)):
            # Attempt to authorize the account.
            local_site_name: (str | None) = None

            if self.local_site:
                local_site_name = self.local_site.name

            password = credentials.get('password')
            two_factor_auth_code = credentials.get('two_factor_auth_code')

            authorize_kwargs: KwargsDict = dict({
                'username': username,
                'password': password,
                'hosting_url': hosting_url,
                'two_factor_auth_code': two_factor_auth_code,
                'local_site_name': local_site_name,
                'credentials': credentials,
            }, **extra_authorize_kwargs)

            reauth = False

            try:
                self.authorize(hosting_account, hosting_service_id,
                               **authorize_kwargs)
            except CertificateVerificationError as e:
                if trust_host:
                    cert = e.certificate

                    # In the common case, we'll have a certificate with
                    # data, but in the event that there's an implementation
                    # error with some of the data, we'll want to log it
                    # and raise.
                    if cert is not None and cert.cert_data:
                        # Add the certificate and attempt re-authorization.
                        cert_manager.add_certificate(cert)
                        reauth = True
                    else:
                        # There's an implementation error. Log the details
                        # and let the exception raise.
                        if cert is None:
                            logger.error(
                                'Hosting service %r provided a '
                                'CertificateVerificationError without a '
                                'certificate. This is an implementation error '
                                'and must be fixed.',
                                hosting_service_id,
                            )
                        elif not cert.cert_data:
                            logger.error(
                                'Hosting service %r provided a '
                                'CertificateVerificationError with a '
                                'certificate but without certificate data. '
                                'This is an implementation error and must '
                                'be fixed.',
                                hosting_service_id,
                            )

                if not reauth:
                    raise
            except LegacyUnverifiedCertificateError as e:
                if trust_host:
                    hosting_account.accept_certificate(e.certificate)
                    reauth = True
                else:
                    raise

            if reauth:
                self.authorize(hosting_account, hosting_service_id,
                               **authorize_kwargs)

        if save:
            hosting_account.save()

        return hosting_account

    def authorize(
        self,
        hosting_account: HostingServiceAccount,
        hosting_service_id: str,
        username: (str | None) = None,
        local_site_name: (str | None) = None,
        **kwargs,
    ) -> None:
        """Authorize the service.

        Args:
            hosting_account (reviewboard.hostingsvcs.models.
                             HostingServiceAccount):
                The hosting service account.

            hosting_service_id (str):
                The ID of the hosting service.

            username (str):
                The username for the account.

            local_site_name (str, optional):
                The Local Site name, if any, that the account should be
                bound to.

            **kwargs (dict):
                Keyword arguments to pass into the service authorize function.
        """
        try:
            hosting_account.service.authorize(username=username,
                                              local_site_name=local_site_name,
                                              **kwargs)
        except TwoFactorAuthCodeRequiredError:
            # Mark this as required for the next form render.
            self.fields['hosting_account_two_factor_auth_code']\
                .required = True

            # Re-raise the error.
            raise
        except AuthorizationError:
            logger.exception('Authorization error linking hosting '
                             'account ID=%r for hosting service=%r, '
                             'username=%r, LocalSite=%r',
                             hosting_account.pk, hosting_service_id,
                             username, local_site_name)

            # Re-raise the error.
            raise
        except (CertificateVerificationError,
                LegacyUnverifiedCertificateError):
            # Re-raise the error so the user will see the "I trust this
            # host" prompt.
            raise
        except Exception:
            logger.exception('Unknown error linking hosting account '
                             'ID=%r for hosting service=%r, '
                             'username=%r, LocalSite=%r',
                             hosting_account.pk, hosting_service_id,
                             username, local_site_name)

            # Re-raise the error.
            raise

    def clean_hosting_url(self) -> str | None:
        """Clean the hosting URL field.

        Returns:
            str:
            A string containing the hosting URL, or ``None``.
        """
        return self.cleaned_data['hosting_url'] or None


class BaseHostingServiceRepositoryForm(_HostingServiceSubFormMixin,
                                       BaseRepositoryInfoSubForm):
    """Base form for collecting information for a hosting service repository.

    This is responsible for providing fields used to communicate with a
    particular hosting service, such as a registered organization name or ID
    on the service. There may be one global form (set in
    :py:attr:`BaseHostingService.form
    <reviewboard.hostingsvcs.base.hosting_service.BaseHostingService.form>` or
    one per plan.

    Each field will be stored directly in :py:attr:`Repository.extra_data
    <reviewboard.scmtools.models.Repository.extra_data>`, using the field's
    name as the key.

    Subclasses are expected to prefix every field with the ID of the hosting
    service, to avoid conflicts.

    Subclasses may also define a ``Meta`` class on the form containing
    ``help_texts`` and ``labels`` attributes, mapping field names to custom
    help text or labels. This is useful if a hosting service has a base form
    for collecting details for each plan, and wants to customize the labels
    and help text for each subclass. For example:

    .. code-block:: python

        from django import forms
        from django.utils.translation import gettext_lazy as _
        from reviewboard.hostingsvcs.base.forms import \
            BaseHostingServiceRepositoryForm


        class MyServiceBaseForm(BaseHostingServiceRepositoryForm):
            myservice_owner = forms.CharField(max_length=64)


        class MyServiceOrgPlanForm(MyServiceBaseForm):
            class Meta:
                labels = {
                    'myservice_owner': _('User'),
                }

                help_texts = {
                    'myservice_owner': _(
                        'The username of the user owning the repository.'
                    ),
                }


        class MyServicePersonalPlanForm(MyServiceBaseForm):
            class Meta:
                labels = {
                    'myservice_owner': _('Organization'),
                }

                help_texts = {
                    'myservice_owner': _(
                        'The ID of the organization owning the repository.'
                    ),
                }

    Version Changed:
        6.0:
        * Moved from :py:mod:`reviewboard.hostingsvcs.forms` to
          :py:mod:`reviewboard.hostingsvcs.base.forms` and renamed from
          ``HostingServiceForm`` to ``BaseHostingServiceRepositoryForm``.
    """

    def get_initial_data(self) -> JSONDict:
        """Return initial data for the form.

        This will load information from the repository's
        :py:attr:`~reviewboard.scmtools.models.Repository.extra_data` into the
        form's fields.

        Returns:
            dict:
            Initial data for the form.
        """
        return self.get_field_data_from(self.repository)

    def load(
        self,
        repository: (Repository | None) = None,
        **kwargs,
    ) -> None:
        """Load information for the form.

        By default, this will populate initial values returned in
        :py:meth:`get_initial_data`. Subclasses can override this to set
        other fields or state as needed.

        Args:
            repository (reviewboard.scmtools.models.Repository, optional):
                The repository being loaded.

                This is scheduled to be deprecated. Subclasses should use the
                :py:attr:`repository` attribute instead.

            **kwargs (dict, unused):
                Unused keyword arguments.
        """
        super().load()

    def save(
        self,
        repository: (Repository | None) = None,
        **kwargs,
    ) -> None:
        """Save information from the form back to the repository.

        This will set each field in the repository's
        :py:attr:`~reviewboard.scmtools.models.Repository.extra_data`.

        Args:
            repository (reviewboard.scmtools.models.Repository, optional):
                The repository being loaded.

                This is scheduled to be deprecated. Subclasses should use the
                :py:attr:`repository` attribute instead.

            **kwargs (dict, unused):
                Unused keyword arguments.
        """
        if repository is None:
            repository = self.repository

            assert repository is not None

        for key, value in self.cleaned_data.items():
            key = self.add_prefix(key)
            repository.extra_data[key] = value



"""Base test case support for extension hook tests."""

from djblets.extensions.manager import ExtensionManager
from djblets.extensions.models import RegisteredExtension
from djblets.extensions.testing.testcases import ExtensionTestCaseMixin

from reviewboard.extensions.base import Extension
from reviewboard.testing.testcase import TestCase


class DummyExtension(Extension):
    registration = RegisteredExtension()


class ExtensionHookTestCaseMixin(ExtensionTestCaseMixin):
    """Mixin for extension hook unit tests.

    Version Added:
        3.0.24
    """

    extension_class = DummyExtension

    def get_extension_manager(self):
        """Return an extension manager for the test.

        The result will always be a new, un-keyed extension manager.

        Returns:
            djblets.extensions.manager.ExtensionManager:
            The new extension manager.
        """
        return ExtensionManager('')


class BaseExtensionHookTestCase(ExtensionHookTestCaseMixin, TestCase):
    """Base class for extension hook unit tests.

    Version Added:
        3.0.24
    """



"""Models for file attachments and related objects."""

from __future__ import annotations

import logging
import os
from inspect import signature
from typing import ClassVar, List, Optional, Sequence, TYPE_CHECKING

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import Max
from django.utils.translation import gettext_lazy as _
from djblets.db.fields import JSONField, RelationCounterField
from djblets.util.decorators import cached_property
from typing_extensions import TypeAlias

from reviewboard.admin.server import build_server_url
from reviewboard.attachments.managers import FileAttachmentManager
from reviewboard.attachments.mimetypes import MimetypeHandler
from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.diffviewer.diffutils import get_sha256
from reviewboard.diffviewer.models import FileDiff
from reviewboard.scmtools.models import Repository
from reviewboard.site.models import LocalSite
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from reviewboard.reviews.models import ReviewRequest


logger = logging.getLogger(__name__)


class FileAttachmentHistory(models.Model):
    """Revision history for a single file attachment.

    This tracks multiple revisions of the same file attachment (for instance,
    when someone replaces a screenshot with an updated version).
    """

    display_position = models.IntegerField()
    latest_revision = RelationCounterField('file_attachments')

    def get_revision_to_id_map(self):
        """Return a map from revision number to FileAttachment ID."""
        results = {}

        for attachment in self.file_attachments.all():
            results[attachment.attachment_revision] = attachment.id

        return results

    @staticmethod
    def compute_next_display_position(review_request):
        """Compute the display position for a new FileAttachmentHistory."""
        # Right now, display_position is monotonically increasing for each
        # review request. In the future this might be extended to allow the
        # user to change the order of attachments on the page.
        max_position = (
            FileAttachmentHistory.objects
            .filter(review_request=review_request)
            .aggregate(Max('display_position'))
            .get('display_position__max')) or 0

        return max_position + 1

    class Meta:
        db_table = 'attachments_fileattachmenthistory'
        verbose_name = _('File Attachment History')
        verbose_name_plural = _('File Attachment Histories')


class FileAttachment(models.Model):
    """A file associated with a review request.

    Like diffs, a file can have comments associated with it.
    These comments are of type
    :py:class:`reviewboard.reviews.models.FileAttachmentComment`.
    """

    caption = models.CharField(_('caption'), max_length=256, blank=True)
    draft_caption = models.CharField(_('draft caption'),
                                     max_length=256, blank=True)
    orig_filename = models.CharField(_('original filename'),
                                     max_length=256, blank=True, null=True)
    user = models.ForeignKey(User,
                             on_delete=models.CASCADE,
                             blank=True,
                             null=True,
                             related_name='file_attachments')

    #: The local site that this attachment is on.
    #:
    #: Prior to RB 8.1, this was only  present for "user" file attachments
    #: (that is, attachments that were uploaded into a comment box or otherwise
    #: created from the user file attachments API) and for ones linked to
    #: file diffs. Attachments which are uploaded directly to a Review Request,
    #: not part of the diff, did not have this relation populated.
    #:
    #: On RB 8.1 and higher we set this for all file attachments. Callers
    #: should still use the :py:meth:`get_local_site` method instead of
    #: accessing this directly for consistency.
    local_site = models.ForeignKey(LocalSite,
                                   on_delete=models.CASCADE,
                                   blank=True,
                                   null=True,
                                   related_name='file_attachments')

    uuid = models.CharField(_('uuid'), max_length=255, blank=True)

    file = models.FileField(_('file'),
                            max_length=512,
                            blank=True,
                            null=True,
                            upload_to=os.path.join('uploaded', 'files',
                                                   '%Y', '%m', '%d'))
    mimetype = models.CharField(_('mimetype'), max_length=256, blank=True)

    extra_data = JSONField(null=True)

    # repo_path, repo_revision, and repository are used to identify
    # FileAttachments associated with committed binary files in a source tree.
    # They are not used for new files that don't yet have a revision.
    #
    # For new files, the added_in_filediff association is used.
    repo_path = models.CharField(_('repository file path'),
                                 max_length=1024,
                                 blank=True,
                                 null=True)
    repo_revision = models.CharField(_('repository file revision'),
                                     max_length=64,
                                     blank=True,
                                     null=True,
                                     db_index=True)
    repository = models.ForeignKey(Repository,
                                   on_delete=models.CASCADE,
                                   blank=True,
                                   null=True,
                                   related_name='file_attachments')
    added_in_filediff = models.ForeignKey(FileDiff,
                                          on_delete=models.CASCADE,
                                          blank=True,
                                          null=True,
                                          related_name='added_attachments')

    attachment_history = models.ForeignKey(FileAttachmentHistory,
                                           on_delete=models.CASCADE,
                                           blank=True,
                                           null=True,
                                           related_name='file_attachments')
    attachment_revision = models.IntegerField(default=0)

    objects: ClassVar[FileAttachmentManager] = FileAttachmentManager()

    #: The extra_data key that denotes if :py:attr:`local_site` has been set.
    #:
    #: Version Added:
    #:     8.1
    DEFINED_LOCAL_SITE_KEY = '__defined_local_site'

    @property
    def mimetype_handler(self):
        """Return the mimetype handler for this file."""
        if not hasattr(self, '_thumbnail'):
            self._thumbnail = MimetypeHandler.for_type(self)

        return self._thumbnail

    @property
    def review_ui(self):
        """Return the review UI for this file."""
        if not hasattr(self, '_review_ui'):
            self._review_ui = None

            from reviewboard.reviews.ui.base import ReviewUI
            review_ui_class = ReviewUI.for_object(self)

            if review_ui_class:
                try:
                    self._review_ui = review_ui_class(
                        obj=self,
                        review_request=self.get_review_request())
                except ObjectDoesNotExist as e:
                    logger.error('Unable to load Review UI %r for %s: %s',
                                 review_ui_class, self, e)
                except Exception as e:
                    logger.exception('Error instantiating Review UI %r" %s',
                                     review_ui_class, e)

        return self._review_ui

    def _get_thumbnail(self):
        """Return the thumbnail for display."""
        if not self.mimetype_handler:
            return None

        try:
            return self.mimetype_handler.get_thumbnail()
        except Exception as e:
            logger.error('Error when calling get_thumbnail for '
                         'MimetypeHandler %r: %s',
                         self.mimetype_handler, e, exc_info=True)
            return None

    def _set_thumbnail(self, data):
        """Set the thumbnail."""
        if not self.mimetype_handler:
            return None

        try:
            self.mimetype_handler.set_thumbnail(data)
        except Exception as e:
            logger.error('Error when calling get_thumbnail for '
                         'MimetypeHandler %r: %s',
                         self.mimetype_handler, e, exc_info=True)
            return None

    thumbnail = property(_get_thumbnail, _set_thumbnail)

    @property
    def filename(self):
        """Return the filename for display purposes."""
        # Older versions of Review Board didn't store the original filename,
        # instead just using the FileField's name. Newer versions have
        # a dedicated filename field.
        if self.file:
            alt = os.path.basename(self.file.name)
        else:
            alt = None

        return self.orig_filename or alt

    @property
    def display_name(self):
        """Return a display name for the file."""
        if self.caption:
            return self.caption
        else:
            return self.filename

    @property
    def icon_url(self):
        """Return the icon URL for this file."""
        if not self.mimetype_handler:
            return None

        try:
            return self.mimetype_handler.get_icon_url()
        except Exception as e:
            logger.error('Error when calling get_thumbnail for '
                         'MimetypeHandler %r: %s',
                         self.mimetype_handler, e, exc_info=True)
            return None

    @property
    def is_from_diff(self):
        """Return if this file attachment is associated with a diff."""
        return (self.repository_id is not None or
                self.added_in_filediff_id is not None)

    @property
    def num_revisions(self):
        """Return the number of revisions of this attachment."""
        return FileAttachment.objects.filter(
            attachment_history=self.attachment_history_id).count() - 1

    @cached_property
    def sha256_checksum(self) -> str:
        """The SHA256 hash of the file.

        Version Added:
            8.0
        """
        checksum: (str | None) = self.extra_data.get('sha256_checksum')

        if checksum is None:
            f = self.file

            try:
                checksum = get_sha256(f.file)
            finally:
                f.close()

            self.extra_data['sha256_checksum'] = checksum
            self.save(update_fields=['extra_data'])

        return checksum

    def __str__(self):
        """Return a string representation of this file for the admin list."""
        return self.caption

    def get_review_request(self) -> ReviewRequest:
        """Return the ReviewRequest that this file is attached to.

        Raises:
            django.core.exceptions.ObjectDoesNotExist:
                The review request does not exist or hasn't been linked yet.
        """
        if hasattr(self, '_review_request'):
            return self._review_request

        try:
            self._review_request = self.review_request.all()[0]
        except IndexError:
            try:
                self._review_request = self.inactive_review_request.all()[0]
            except IndexError:
                # Maybe it's on a draft.
                try:
                    draft = self.drafts.get()
                except ObjectDoesNotExist:
                    draft = self.inactive_drafts.get()

                self._review_request = draft.review_request

        return self._review_request

    def get_local_site(self) -> LocalSite | None:
        """Return the local site for this attachment.

        Version Changed:
            8.1:
            This will return :py:attr:`local_site` if we know that a value has
            been explicitly set for it. Otherwise, this will determine the
            local site, save that value to :py:attr:`local_site` and return
            it.

        Version Added:
            7.0.3

        Returns:
            reviewboard.site.models.LocalSite:
            The local site that this attachment is related to, if any.
        """
        local_site = self.local_site
        defined_local_site_key = self.DEFINED_LOCAL_SITE_KEY

        if self.extra_data.get(defined_local_site_key, False):
            # We have an explicit value for the local site (passed on creation
            # for RB 8.1+ or we figured it out below already). We can use it.
            return local_site

        # Pre RB 8.1, we need to figure out the local site and save it.
        if self.user_id is None and local_site is None:
            # This could be a file attachment that doesn't have a local site,
            # or one that relies on its review request for the local site.
            try:
                local_site = self.get_review_request().local_site
            except ObjectDoesNotExist:
                # We're being asked for the local site before the attachment
                # has been linked to a review request. This shouldn't happen
                # in normal operation, RB 8.1+ sets local_site at creation
                # and any existing attachments will have already been linked
                # to their review request right after their first save.
                #
                # Fall back to ``None`` and log just in case.
                logger.error(
                    'Could not determine Local Site for file attachment %s. '
                    'Set either the Local Site or the review request on the '
                    'attachment to fix this.',
                    self.pk)

                return None

        # We now know the local_site value for pre RB 8.1 file attachments.
        # Save it so we don't have to figure it out again on the next access.
        self.local_site = local_site
        self.extra_data[defined_local_site_key] = True

        if self.pk is not None:
            self.save(update_fields=['local_site', 'extra_data'])

        return local_site

    def get_comments(self):
        """Return all the comments made on this file attachment."""
        if not hasattr(self, '_comments'):
            self._comments = list(self.comments.all())

        return self._comments

    def get_raw_download_url(self) -> Optional[str]:
        """Return the absolute URL to download this file.

        The URL will be determined by the storage backend. It may be
        accessible for only a limited amount of time, and may or may not be
        cacheable by the browser.

        Version Added:
            7.0

        Returns:
            str:
            The absolute URL to the file.

            This will be ``None`` if there's no file backing for any reason.
        """
        if not self.file:
            return None

        url = self.file.url

        if not url or url.startswith(('http:', 'https:')):
            return url

        return build_server_url(url)

    def get_raw_thumbnail_image_url(
        self,
        *,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ) -> Optional[str]:
        """Return the absolute URL for an image thumbnail for this file.

        The URL will be determined by the storage backend. It may be
        accessible for only a limited amount of time, and may or may not be
        cacheable by the browser.

        Not all file attachments support image thumbnails. If not supported,
        this will be ``None``.

        Version Added:
            7.0

        Returns:
            str:
            The absolute URL to the file.

            This will be ``None`` if there's no file backing for any reason,
            or if the attachment doesn't support image thumbnails.
        """
        url: Optional[str] = None
        mimetype_handler = self.mimetype_handler

        if mimetype_handler:
            try:
                url = mimetype_handler.get_raw_thumbnail_image_url(
                    width=width,
                    height=height)
            except NotImplementedError:
                # This file type doesn't support image thumbnails.
                pass

        if not url or url.startswith(('http:', 'https:')):
            return url

        return build_server_url(url)

    def get_absolute_url(self) -> Optional[str]:
        """Return the absolute URL for accessing the file.

        This will return the correct URL for either user-uploaded or
        review request file attachments.

        If the association could not be determined, this will return ``None``.

        The URL will always be a full absolute URL, usable in e-mails and
        other sources.

        Returns:
            str:
            The URL to access the file attachment contents.
        """
        if self.user is not None:
            # This is a user-uploaded file attachment.
            path = local_site_reverse(
                'user-file-attachment',
                local_site=self.get_local_site(),
                kwargs={
                    'file_attachment_uuid': self.uuid,
                    'username': self.user.username,
                })
        else:
            # This is a file attachment on a review request.
            try:
                review_request = self.get_review_request()

                path = local_site_reverse(
                    'download-file-attachment',
                    local_site=self.get_local_site(),
                    kwargs={
                        'file_attachment_id': self.pk,
                        'review_request_id': review_request.get_display_id(),
                    })
            except ObjectDoesNotExist:
                return None

        return build_server_url(path)

    def is_review_ui_accessible_by(
        self,
        user: User,
    ) -> bool:
        """Return whether a user can access the file attachment's review UI.

        This will check that a review UI exists for the file attachment and
        that it's enabled for the provided user and review request.

        Version Added:
            7.0.3

        Args:
            user (django.contrib.auth.models.User):
                The user who is accessing the review UI.

        Returns:
            bool:
            ``True`` if a review UI exists and can be accessed by the user.
            ``False`` if the review UI does not exist, cannot be used, or
            there's an error when checking.
        """
        review_ui = self.review_ui

        if not review_ui:
            return False

        review_request = self.get_review_request()

        try:
            params = signature(review_ui.is_enabled_for).parameters

            if 'file_attachment' in params:
                RemovedInReviewBoard90Warning.warn(
                    'The file_attachment parameter to ReviewUI.is_enabled_for '
                    'has been removed. Please use obj= instead in Review UI %r'
                    % review_ui)

                return review_ui.is_enabled_for(
                    user=user,
                    review_request=review_request,
                    file_attachment=self)
            else:
                return review_ui.is_enabled_for(
                    user=user,
                    review_request=review_request,
                    obj=self)
        except Exception as e:
            logger.exception('Error when calling is_enabled_for with '
                             'ReviewUI %r: %s',
                             review_ui, e)
            return False

    def is_accessible_by(self, user):
        """Returns whether or not the user has access to this FileAttachment.

        This checks that the user has access to the LocalSite if the attachment
        is associated with a local site. This is only applicable for user owned
        file attachments.
        """
        return (self.user_id is not None and
                user.is_authenticated and
                (user.is_superuser or self.user_id == user.pk) and
                (not self.local_site or
                 self.local_site.is_accessible_by(user)))

    def is_mutable_by(self, user):
        """Returns whether or not a user can modify this FileAttachment.

        This checks that the user is either a superuser or the owner of the
        file attachment. This is only applicable for user owned file
        attachments.
        """
        return (self.user_id is not None and
                user.is_authenticated and
                (user.is_superuser or self.user_id == user.pk))

    class Meta:
        db_table = 'attachments_fileattachment'
        get_latest_by = 'attachment_revision'
        verbose_name = _('File Attachment')
        verbose_name_plural = _('File Attachments')


def get_latest_file_attachments(
    file_attachments: FileAttachmentSequence,
) -> List[FileAttachment]:
    """Filter the list of file attachments to only return the latest revisions.

    Args:
        file_attachments (list of
                          reviewboard.attachments.models.FileAttachment):
            The file attachments to filter.

    Returns:
        list of reviewboard.attachments.models.FileAttachment:
        The list of file attachments that are the latest revisions in their
        respective histories.
    """
    file_attachment_histories = FileAttachmentHistory.objects.filter(
        file_attachments__in=file_attachments)
    latest = {
        data['id']: data['latest_revision']
        for data in file_attachment_histories.values('id', 'latest_revision')
    }

    return [
        f
        for f in file_attachments
        if (not f.is_from_diff and
            f.attachment_revision == latest[f.attachment_history_id])
    ]


#: Type alias for a sequence of file attachments.
#:
#: Version Added:
#:     6.0
FileAttachmentSequence: TypeAlias = Sequence[FileAttachment]



"""Unit tests for reviewboard.attachments.models.FileAttachment.

Version Added:
    7.0.3:
    This was split off from :py:mod:`reviewboard.attachments.tests`.
"""

from __future__ import annotations

import json
import os
from datetime import datetime

import kgb
from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import User
from django.core.files.uploadedfile import SimpleUploadedFile
from django.db.models import Q
from djblets.testing.decorators import add_fixtures

from reviewboard.attachments.forms import UploadFileForm
from reviewboard.attachments.models import (FileAttachment,
                                            FileAttachmentHistory)
from reviewboard.attachments.tests.base import BaseFileAttachmentTestCase
from reviewboard.reviews.models import ReviewRequest, ReviewRequestDraft
from reviewboard.reviews.ui.image import ImageReviewUI
from reviewboard.site.models import LocalSite


class FileAttachmentTests(kgb.SpyAgency, BaseFileAttachmentTestCase):
    """Unit tests for reviewboard.attachments.models.FileAttachment.

    Version Added:
        7.0.3:
        This was split off from :py:mod:`reviewboard.attachments.tests`.
    """

    @add_fixtures(['test_users', 'test_scmtools'])
    def test_upload_file(self):
        """Testing uploading a file attachment"""
        review_request = self.create_review_request(publish=True)

        file = self.make_uploaded_file()
        form = UploadFileForm(review_request, files={
            'path': file,
        })
        self.assertTrue(form.is_valid())

        file_attachment = form.create()
        file_attachment.refresh_from_db()

        self.assertTrue(os.path.basename(file_attachment.file.name).endswith(
            '__logo.png'))
        self.assertEqual(file_attachment.mimetype, 'image/png')
        self.assertIsNone(file_attachment.get_local_site())
        self.assertEqual(
            file_attachment.extra_data,
            {
                self.DEFINED_LOCAL_SITE_KEY: True,
                'sha256_checksum': ('1931a3b367e2913d28f9587dbd0ccf79b2c'
                                    '2225de7c47550dd1cc49085077e49'),
            })

    @add_fixtures(['test_users', 'test_scmtools'])
    def test_upload_file_with_history(self):
        """Testing uploading a file attachment to an existing
        FileAttachmentHistory
        """
        review_request_1 = self.create_review_request(publish=True)
        history = FileAttachmentHistory.objects.create(display_position=0)
        review_request_1.file_attachment_histories.add(history)

        file = self.make_uploaded_file()
        form = UploadFileForm(review_request_1,
                              data={'attachment_history': history.pk},
                              files={'path': file})
        self.assertTrue(form.is_valid())
        form.create()

    @add_fixtures(['test_users', 'test_scmtools'])
    def test_upload_file_with_history_mismatch(self):
        """Testing uploading a file attachment to an existing
        FileAttachmentHistory with a mismatched review request
        """
        review_request_1 = self.create_review_request(publish=True)
        review_request_2 = self.create_review_request(publish=True)
        uploaded_file = self.make_uploaded_file()

        history = FileAttachmentHistory.objects.create(display_position=0)
        review_request_1.file_attachment_histories.add(history)

        form = UploadFileForm(review_request_2,
                              data={'attachment_history': history.pk},
                              files={'path': uploaded_file})
        self.assertFalse(form.is_valid())

    @add_fixtures(['test_users', 'test_scmtools'])
    def test_upload_file_revisions(self):
        """Testing uploading multiple revisions of a file"""
        user = User.objects.create_user(username='testuser')
        review_request = self.create_review_request(publish=True,
                                                    target_people=[user])
        history = FileAttachmentHistory.objects.create(display_position=0)
        review_request.file_attachment_histories.add(history)
        uploaded_file = self.make_uploaded_file()

        # Add a file with the given history
        form = UploadFileForm(review_request,
                              data={'attachment_history': history.pk},
                              files={'path': uploaded_file})
        self.assertTrue(form.is_valid())
        file_attachment = form.create()
        history = FileAttachmentHistory.objects.get(pk=history.pk)
        self.assertEqual(file_attachment.attachment_revision, 1)
        self.assertEqual(history.latest_revision, 1)
        self.assertEqual(history.display_position, 0)

        review_request.get_draft().publish()
        # Post an update
        form = UploadFileForm(review_request,
                              data={'attachment_history': history.pk},
                              files={'path': uploaded_file})
        self.assertTrue(form.is_valid())
        file_attachment = form.create()
        history = FileAttachmentHistory.objects.get(pk=history.pk)
        self.assertEqual(file_attachment.attachment_revision, 2)
        self.assertEqual(history.latest_revision, 2)
        self.assertEqual(history.display_position, 0)

        review_request.get_draft().publish()

        # Post two updates without publishing the draft in between
        form = UploadFileForm(review_request,
                              data={'attachment_history': history.pk},
                              files={'path': uploaded_file})
        self.assertTrue(form.is_valid())
        file_attachment = form.create()
        history = FileAttachmentHistory.objects.get(pk=history.pk)
        self.assertEqual(file_attachment.attachment_revision, 3)
        self.assertEqual(history.latest_revision, 3)
        self.assertEqual(history.display_position, 0)

        form = UploadFileForm(review_request,
                              data={'attachment_history': history.pk},
                              files={'path': uploaded_file})
        self.assertTrue(form.is_valid())
        file_attachment = form.create()
        history = FileAttachmentHistory.objects.get(pk=history.pk)
        self.assertEqual(file_attachment.attachment_revision, 3)
        self.assertEqual(history.latest_revision, 3)
        self.assertEqual(history.display_position, 0)

        # Add another (unrelated) file to check display position
        form = UploadFileForm(review_request,
                              files={'path': uploaded_file})
        self.assertTrue(form.is_valid())
        file_attachment = form.create()
        self.assertEqual(file_attachment.attachment_revision, 1)
        self.assertEqual(file_attachment.attachment_history.latest_revision, 1)
        self.assertEqual(file_attachment.attachment_history.display_position,
                         1)

    @add_fixtures(['test_users', 'test_scmtools'])
    def test_upload_file_with_extra_data(self):
        """Testing uploading a file attachment with extra data"""
        class TestObject():
            def to_json(self):
                return {
                    'foo': 'bar'
                }

        review_request = self.create_review_request(publish=True)

        file = self.make_uploaded_file()
        form = UploadFileForm(
            review_request,
            data={
                'extra_data': {
                    'test_bool': True,
                    'test_date': datetime(2023, 1, 26, 5, 30, 3, 123456),
                    'test_int': 1,
                    'test_list': [1, 2, 3],
                    'test_nested_dict': {
                        'foo': 2,
                        'bar': 'baz',
                    },
                    'test_none': None,
                    'test_obj': TestObject(),
                    'test_str': 'test',
                }
            },
            files={'path': file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create()
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.extra_data, {
            self.DEFINED_LOCAL_SITE_KEY: True,
            'sha256_checksum': ('1931a3b367e2913d28f9587dbd0ccf79b2c'
                                '2225de7c47550dd1cc49085077e49'),
            'test_bool': True,
            'test_date': '2023-01-26T05:30:03.123',
            'test_int': 1,
            'test_list': [1, 2, 3],
            'test_nested_dict': {
                'foo': 2,
                'bar': 'baz',
            },
            'test_none': None,
            'test_obj': {
                'foo': 'bar',
            },
            'test_str': 'test',
        })

    @add_fixtures(['test_users', 'test_scmtools'])
    def test_upload_file_with_extra_data_string(self):
        """Testing uploading a file attachment with extra data passed as a
        JSON string
        """
        review_request = self.create_review_request(publish=True)

        file = self.make_uploaded_file()
        form = UploadFileForm(
            review_request,
            data={
                'extra_data': json.dumps({
                    'test_bool': True,
                    'test_int': 1,
                    'test_list': [1, 2, 3],
                    'test_nested_dict': {
                        'foo': 2,
                        'bar': 'baz',
                    },
                    'test_none': None,
                    'test_str': 'test',
                })
            },
            files={'path': file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create()
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.extra_data, {
            self.DEFINED_LOCAL_SITE_KEY: True,
            'sha256_checksum': ('1931a3b367e2913d28f9587dbd0ccf79b2c'
                                '2225de7c47550dd1cc49085077e49'),
            'test_bool': True,
            'test_int': 1,
            'test_list': [1, 2, 3],
            'test_nested_dict': {
                'foo': 2,
                'bar': 'baz',
            },
            'test_none': None,
            'test_str': 'test',
        })

    @add_fixtures(['test_users', 'test_scmtools'])
    def test_upload_file_with_extra_data_empties(self):
        """Testing uploading a file attachment with extra data that contains
        empty values
        """
        review_request = self.create_review_request(publish=True)
        file = self.make_uploaded_file()

        form = UploadFileForm(
            review_request,
            data={
                'extra_data': {}
            },
            files={'path': file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create()
        file_attachment.refresh_from_db()
        self.assertEqual(
            file_attachment.extra_data,
            {
                self.DEFINED_LOCAL_SITE_KEY: True,
                'sha256_checksum': ('1931a3b367e2913d28f9587dbd0ccf79b2c'
                                    '2225de7c47550dd1cc49085077e49'),
            })

        form = UploadFileForm(
            review_request,
            data={
                'extra_data': json.dumps(None)
            },
            files={'path': file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create()
        file_attachment.refresh_from_db()
        self.assertEqual(
            file_attachment.extra_data,
            {
                self.DEFINED_LOCAL_SITE_KEY: True,
                'sha256_checksum': ('1931a3b367e2913d28f9587dbd0ccf79b2c'
                                    '2225de7c47550dd1cc49085077e49'),
            })

        form = UploadFileForm(
            review_request,
            data={
                'extra_data': None
            },
            files={'path': file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create()
        file_attachment.refresh_from_db()
        self.assertEqual(
            file_attachment.extra_data,
            {
                self.DEFINED_LOCAL_SITE_KEY: True,
                'sha256_checksum': ('1931a3b367e2913d28f9587dbd0ccf79b2c'
                                    '2225de7c47550dd1cc49085077e49'),
            })

        form = UploadFileForm(
            review_request,
            data={
                'extra_data': {
                    'test_list': [],
                    'test_nested_dict': {},
                    'test_none': None,
                    'test_str': '',
                }
            },
            files={'path': file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create()
        file_attachment.refresh_from_db()
        self.assertEqual(file_attachment.extra_data, {
            self.DEFINED_LOCAL_SITE_KEY: True,
            'sha256_checksum': ('1931a3b367e2913d28f9587dbd0ccf79b2c'
                                '2225de7c47550dd1cc49085077e49'),
            'test_list': [],
            'test_nested_dict': {},
            'test_none': None,
            'test_str': '',
        })

    def test_upload_file_sets_local_site(self) -> None:
        """Testing uploading a file attachment sets local_site when attached
        to a review request without a local site
        """
        review_request = self.create_review_request()

        form = UploadFileForm(review_request,
                              files={'path': self.make_uploaded_file()})
        self.assertTrue(form.is_valid())

        file_attachment = form.create()
        file_attachment.refresh_from_db()

        self.assertIsNone(file_attachment.local_site)
        self.assertTrue(
            file_attachment.extra_data[self.DEFINED_LOCAL_SITE_KEY])

    def test_upload_file_sets_local_site_with_local_site(self) -> None:
        """Testing uploading a file attachment sets local_site when attached
        to a review request with a local site
        """
        review_request = self.create_review_request(with_local_site=True)

        form = UploadFileForm(review_request,
                              files={'path': self.make_uploaded_file()})
        self.assertTrue(form.is_valid())

        file_attachment = form.create()
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.local_site,
                         review_request.local_site)
        self.assertTrue(
            file_attachment.extra_data[self.DEFINED_LOCAL_SITE_KEY])

    def test_is_from_diff_with_no_association(self):
        """Testing FileAttachment.is_from_diff with standard attachment"""
        file_attachment = FileAttachment()

        self.assertFalse(file_attachment.is_from_diff)

    @add_fixtures(['test_scmtools'])
    def test_is_from_diff_with_repository(self):
        """Testing FileAttachment.is_from_diff with repository association"""
        repository = self.create_repository()
        file_attachment = FileAttachment(repository=repository)

        self.assertTrue(file_attachment.is_from_diff)

    @add_fixtures(['test_scmtools'])
    def test_is_from_diff_with_filediff(self):
        """Testing FileAttachment.is_from_diff with filediff association"""
        filediff = self.make_filediff()
        file_attachment = FileAttachment(added_in_filediff=filediff)

        self.assertTrue(file_attachment.is_from_diff)

    @add_fixtures(['test_users', 'test_scmtools'])
    def test_utf16_thumbnail(self):
        """Testing file attachment thumbnail generation for UTF-16 files"""
        filename = os.path.join(os.path.dirname(__file__),
                                'testdata', 'utf-16.txt')
        with open(filename, 'rb') as f:
            review_request = self.create_review_request(publish=True)

            file = SimpleUploadedFile(
                f.name,
                f.read(),
                content_type='text/plain;charset=utf-16le')
            form = UploadFileForm(review_request, files={'path': file})
            form.is_valid()

            file_attachment = form.create()

            self.assertEqual(
                file_attachment.thumbnail,
                '<div class="file-thumbnail"> <div class="file-thumbnail-clipp'
                'ed"><pre>UTF-16le encoded sample plain-text file</pre><pre>'
                '\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e'
                '\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e'
                '\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e'
                '\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e\u203e</pre>'
                '<pre></pre><pre>Markus Kuhn [\u02c8ma\u02b3k\u028as ku\u02d0'
                'n] &lt;http://www.cl.cam.ac.uk/~mgk25/&gt; \u2014 2002-07-25'
                '</pre><pre></pre><pre></pre><pre>The ASCII compatible UTF-8 '
                'encoding used in this plain-text file</pre><pre>is defined '
                'in Unicode, ISO 10646-1, and RFC 2279.</pre><pre></pre><pre>'
                '</pre><pre>Using Unicode/UTF-8, you can write in emails and '
                'source code things such as</pre><pre></pre><pre>Mathematics '
                'and sciences:</pre><pre></pre><pre>  \u222e E\u22c5da = Q,  '
                'n \u2192 \u221e, \u2211 f(i) = \u220f g(i),      \u23a7\u23a1'
                '\u239b\u250c\u2500\u2500\u2500\u2500\u2500\u2510\u239e\u23a4'
                '\u23ab</pre><pre>                                           '
                ' \u23aa\u23a2\u239c\u2502a\xb2+b\xb3 \u239f\u23a5\u23aa'
                '</pre><pre>  \u2200x\u2208</pre></div></div>')

    @add_fixtures(['test_users'])
    def test_is_review_ui_accessible_by_true(self) -> None:
        """Testing FileAttachment.is_review_ui_accessible_by with a user who
        can access the review UI
        """
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(submitter=user)
        file_attachment = self.create_file_attachment(
            review_request=review_request,
            mimetype='image/png')

        self.spy_on(ImageReviewUI.is_enabled_for, op=kgb.SpyOpReturn(True))

        self.assertTrue(file_attachment.is_review_ui_accessible_by(user=user))

    @add_fixtures(['test_users'])
    def test_is_review_ui_accessible_by_false(self) -> None:
        """Testing FileAttachment.is_review_ui_accessible_by with a user who
        cannot access the review UI
        """
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(submitter=user)
        file_attachment = self.create_file_attachment(
            review_request=review_request,
            mimetype='image/png')

        self.spy_on(ImageReviewUI.is_enabled_for, op=kgb.SpyOpReturn(False))

        self.assertFalse(file_attachment.is_review_ui_accessible_by(user=user))

    def test_get_absolute_url(self) -> None:
        """Testing FileAttachment.get_absolute_url"""
        review_request = self.create_review_request()
        file_attachment = self.create_file_attachment(
            review_request=review_request,
            mimetype='image/png')

        self.assertEqual(
            file_attachment.get_absolute_url(),
            'http://example.com/r/1/file/1/download/')

    def test_get_absolute_url_with_local_site(self) -> None:
        """Testing FileAttachment.get_absolute_url with a local site"""
        review_request = self.create_review_request(with_local_site=True)
        local_site_name = review_request.local_site.name
        file_attachment = self.create_file_attachment(
            review_request=review_request,
            mimetype='image/png')

        self.assertEqual(
            file_attachment.get_absolute_url(),
            f'http://example.com/s/{local_site_name}/r/1001/file/1/download/')

    def test_get_local_site(self) -> None:
        """Testing FileAttachment.get_local_site returns the stored local site
        """
        review_request = self.create_review_request(with_local_site=True)
        file_attachment = self.create_file_attachment(
            review_request=review_request)

        self.spy_on(file_attachment.get_review_request)

        with self.assertNumQueries(0):
            local_site = file_attachment.get_local_site()

        self.assertEqual(local_site, review_request.local_site)
        self.assertIsNotNone(local_site)
        self.assertSpyNotCalled(file_attachment.get_review_request)

    def test_get_local_site_none(self) -> None:
        """Testing FileAttachment.get_local_site with local site set to None"""
        review_request = self.create_review_request()
        file_attachment = self.create_file_attachment(
            review_request=review_request)

        with self.assertNumQueries(0):
            local_site = file_attachment.get_local_site()

        self.assertIsNone(local_site)

    def test_get_local_site_sets_local_site(self) -> None:
        """Testing FileAttachment.get_local_site resolves and saves for a
        pre-8.1 attachment on a local site
        """
        review_request = self.create_review_request(with_local_site=True)
        file_attachment = self.create_file_attachment(
            review_request=review_request)
        attachment_pk = file_attachment.pk

        try:
            delattr(file_attachment, '_review_request')
        except AttributeError:
            pass

        file_attachment.local_site = None
        file_attachment.extra_data.pop(self.DEFINED_LOCAL_SITE_KEY, None)
        file_attachment.save(update_fields=['local_site', 'extra_data'])

        # Get and save the local site for an attachment.
        #
        # 3 queries:
        #
        # 1. Fetch the review request.
        # 2. Fetch the local site on the review request.
        # 3. Save the local site on the file attachment.
        queries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'limit': 1,
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments'
                },
                'where': Q(file_attachments__id=attachment_pk),
            },
            {
                'model': LocalSite,
                'tables': {
                    'site_localsite',
                },
                'where': Q(id=self.local_site_id),
            },
            {
                'model': FileAttachment,
                'tables': {
                    'attachments_fileattachment',
                },
                'type': 'UPDATE',
                'where': Q(pk=attachment_pk),
            },
        ]

        with assert_queries(queries):
            local_site = file_attachment.get_local_site()

        self.assertEqual(local_site, review_request.local_site)
        self.assertIsNotNone(local_site)
        self.assertTrue(
            file_attachment.extra_data[self.DEFINED_LOCAL_SITE_KEY])

        with self.assertNumQueries(0):
            # Calling this again shouldn't perform any queries.
            file_attachment.get_local_site()

    def test_get_local_site_sets_local_site_none(self) -> None:
        """Testing FileAttachment.get_local_site resolves and saves for a
        pre-8.1 attachment not on a local site
        """
        review_request = self.create_review_request()
        file_attachment = self.create_file_attachment(
            review_request=review_request)
        attachment_pk = file_attachment.pk

        try:
            delattr(file_attachment, '_review_request')
        except AttributeError:
            pass

        file_attachment.local_site = None
        file_attachment.extra_data.pop(self.DEFINED_LOCAL_SITE_KEY, None)
        file_attachment.save(update_fields=['local_site', 'extra_data'])

        # Get and save the local site for an attachment.
        #
        # 2 queries:
        #
        # 1. Fetch the review request.
        # 2. Save the local site on the file attachment.
        queries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'limit': 1,
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments'
                },
                'where': Q(file_attachments__id=attachment_pk),
            },
            {
                'model': FileAttachment,
                'tables': {
                    'attachments_fileattachment',
                },
                'type': 'UPDATE',
                'where': Q(pk=attachment_pk),
            },
        ]

        with assert_queries(queries):
            local_site = file_attachment.get_local_site()

        self.assertEqual(local_site, review_request.local_site)
        self.assertIsNone(local_site)
        self.assertTrue(
            file_attachment.extra_data[self.DEFINED_LOCAL_SITE_KEY])

        with self.assertNumQueries(0):
            # Calling this again shouldn't perform any queries.
            file_attachment.get_local_site()

    def test_get_local_site_before_linked_review_request(self) -> None:
        """Testing FileAttachment.get_local_site when called before the
        attachment is linked to a review request
        """
        file_attachment = FileAttachment.objects.create(
            extra_data={})
        attachment_pk = file_attachment.pk

        # Get the local site for an attachment.
        #
        # 4 queries:
        #
        # 1. Try to fetch the review request.
        # 2. Try to fetch the inactive review request.
        # 3. Try to fetch the draft review request.
        # 4. Try to fetch the inactive draft review request.
        queries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'limit': 1,
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments'
                },
                'where': Q(file_attachments__id=attachment_pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_inactive_file_attachments':
                        'INNER JOIN',
                },
                'limit': 1,
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_inactive_file_attachments'
                },
                'where': Q(inactive_file_attachments__id=attachment_pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'limit': 21,
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_file_attachments'
                },
                'where': Q(file_attachments__id=attachment_pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'INNER JOIN',
                },
                'limit': 21,
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_inactive_file_attachments'
                },
                'where': Q(inactive_file_attachments__id=attachment_pk),
            },
        ]

        with self.assertLogs() as logs:
            with assert_queries(queries):
                self.assertIsNone(file_attachment.get_local_site())

        self.assertEqual(len(logs.records), 1)
        self.assertEqual(
            logs.records[0].getMessage(),
            'Could not determine Local Site for file attachment 1. '
            'Set either the Local Site or the review request on the '
            'attachment to fix this.')

        # Now link it to a review request on a local site.
        review_request = self.create_review_request(with_local_site=True)
        review_request.file_attachments.add(file_attachment)

        self.assertEqual(file_attachment.get_local_site(),
                         review_request.local_site)
        self.assertTrue(
            file_attachment.extra_data[self.DEFINED_LOCAL_SITE_KEY])



"""The resource for a draft DiffSet."""

from __future__ import annotations

from django.core.exceptions import ObjectDoesNotExist, ValidationError
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_response_errors,
                                       webapi_request_fields)
from djblets.webapi.errors import (DOES_NOT_EXIST,
                                   INVALID_ATTRIBUTE,
                                   INVALID_FORM_DATA,
                                   NOT_LOGGED_IN,
                                   PERMISSION_DENIED)
from djblets.webapi.fields import (BooleanFieldType,
                                   FileFieldType,
                                   StringFieldType)

from reviewboard.diffviewer.commit_utils import deserialize_validation_info
from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.reviews.signals import review_request_diffset_uploaded
from reviewboard.webapi.base import ImportExtraDataError
from reviewboard.webapi.decorators import webapi_check_local_site
from reviewboard.webapi.resources import resources
from reviewboard.webapi.resources.diff import DiffResource


class DraftDiffResource(DiffResource):
    """Provides information on pending draft diffs for a review request.

    This list will only ever contain a maximum of one diff in current
    versions. This is to preserve compatibility with the public
    :ref:`webapi2.0-diff-resource`.

    POSTing to this resource will create or update a review request draft
    with the provided diff. This also mirrors the public diff resource.
    """
    added_in = '2.0'

    name = 'draft_diff'
    uri_name = 'diffs'
    model_parent_key = 'review_request_draft'
    item_result_key = 'diff'
    list_result_key = 'diffs'
    mimetype_list_resource_name = 'diffs'
    mimetype_item_resource_name = 'diff'

    item_child_resources = [
        resources.draft_diffcommit,
        resources.draft_filediff,
    ]

    def get_parent_object(self, diffset):
        return diffset.review_request_draft.get()

    def has_access_permissions(self, request, diffset, *args, **kwargs):
        return diffset.review_request_draft.get().is_accessible_by(
            request.user)

    def get_queryset(self, request, *args, **kwargs):
        try:
            draft = resources.review_request_draft.get_object(
                request, *args, **kwargs)
        except ObjectDoesNotExist:
            raise self.model.DoesNotExist

        return self.model.objects.filter(review_request_draft=draft)

    @webapi_login_required
    @augment_method_from(DiffResource)
    def get_list(self, *args, **kwargs):
        """Returns the list of draft diffs on the review request.

        Each diff has the target revision and list of per-file diffs
        associated with it.
        """
        pass

    @webapi_login_required
    @webapi_check_local_site
    @webapi_response_errors(DOES_NOT_EXIST, INVALID_ATTRIBUTE,
                            INVALID_FORM_DATA, NOT_LOGGED_IN,
                            PERMISSION_DENIED)
    @webapi_request_fields(
        optional={
            'cumulative_diff': {
                'type': FileFieldType,
                'description': (
                    'The cumulative diff of the entire commit series this '
                    'resource represents.'
                ),
            },
            'finalize_commit_series': {
                'type': BooleanFieldType,
                'description': (
                    'Whether or not this is a request to finalize '
                    'the commit series represented by this resource.\n'
                    '\n'
                    'If this is set to ``true``, then both the '
                    '``cumulative_diff`` and ``validation_info`` fields are '
                    'required.'
                ),
            },
            'parent_diff': {
                'type': FileFieldType,
                'description': (
                    'The parent diff of the cumulative diff of the entire '
                    'commit series this resource represents.'
                ),
            },
            'validation_info': {
                'type': StringFieldType,
                'description': (
                    'Validation information returned when validating the last '
                    'commit in the series with the :ref:`DiffCommit '
                    'validation resource '
                    '<webapi2.0-validate-diff-commit-resource>`.'
                ),
            },
        },
        allow_unknown=True
    )
    def update(self, request, finalize_commit_series=False,
               validation_info=None, extra_fields={}, *args, **kwargs):
        """Update a diff.

        This is used for two purposes:

        1. For updating extra data on a draft diff.

           Extra data can be stored later lookup. See
           :ref:`webapi2.0-extra-data` for more information.

        2. For finalization of a draft diff on a review request created with
           commit history.
        """
        try:
            review_request = resources.review_request.get_object(
                request, *args, **kwargs)
            diffset = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not review_request.is_mutable_by(request.user):
            return self.get_no_access_error(request)

        if extra_fields:
            try:
                self.import_extra_data(diffset, diffset.extra_data,
                                       extra_fields)
            except ImportExtraDataError as e:
                return e.error_payload

        if finalize_commit_series:
            if review_request.created_with_history:
                cumulative_diff = request.FILES.get('cumulative_diff')
                parent_diff = request.FILES.get('parent_diff')

                error_rsp = self._finalize_commit_series(request,
                                                         diffset,
                                                         cumulative_diff,
                                                         parent_diff,
                                                         validation_info)

                if error_rsp is not None:
                    return error_rsp

                review_request_draft = diffset.review_request_draft.get()

                # Only add default reviewers if this is the first time we've
                # published any diffsets.
                if review_request.can_add_default_reviewers():
                    review_request_draft.add_default_reviewers()

                # Now that the commit series for the diffset is finalized,
                # emit a signal that a diffset has been uploaded.
                review_request_diffset_uploaded.send(
                    sender=self.__class__,
                    diffset=diffset,
                    review_request_draft=review_request_draft)
            elif dvcs_feature.is_enabled(request=request):
                return INVALID_ATTRIBUTE, {
                    'reason': 'This review request was not created with '
                              'commit history support.',
                }
            else:
                # Otherwise we silently ignore this option.
                finalize_commit_series = False

        if extra_fields or finalize_commit_series:
            diffset.save(update_fields=('extra_data',))

        return 200, {
            self.item_result_key: diffset,
        }

    def _finalize_commit_series(self, request, diffset, cumulative_diff,
                                parent_diff, validation_info):

        """Finalize the commit series represented by the given diffset.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            diffset (reviewboard.diffviewer.models.diffset.DiffSet):
                The diffset representing a commit series.

            cumulative_diff (django.core.files.uploadedfile.UploadedFile):
                The cumulative diff of the entire commit series.

            parent_diff (django.core.files.uploadedfile.UploadedFile):
                The parent diff, if any.

            validation_info (unicode):
                Validation information from the :py:class:`~reviewboard.webapi.
                resources.validate_diffcommit.ValidateDiffCommitResource`.

        Returns:
            tuple:
            If the finalization process is successful, ``None`` will be
            returned. Otherwise, a 2-tuple of the following will be returned:

            * The WebAPI error (:py:class:`~djblets.webapi.errors.
              WebAPIError`).
            * A response to serialize with the error (:py:class:`dict`).
        """
        field_errors = {}

        if cumulative_diff is None:
            field_errors['cumulative_diff'] = [
                'This field is required when finalize_commit_series is set.',
            ]

        if validation_info is None or not validation_info.strip():
            field_errors['validation_info'] = [
                'This field is required when finalize_commit_series is set.',
            ]
        else:
            try:
                validation_info = deserialize_validation_info(validation_info)
            except (TypeError, ValueError) as e:
                field_errors['validation_info'] = [
                    'Could not parse field: %s' % e,
                ]

        if diffset.is_commit_series_finalized:
            return INVALID_ATTRIBUTE, {
                'reason': 'This diff is already finalized.',
            }

        if field_errors:
            return INVALID_FORM_DATA, {
                'fields': field_errors,
            }

        if parent_diff:
            parent_diff_file_contents = parent_diff.read()
        else:
            parent_diff_file_contents = None

        diff_file_contents = cumulative_diff.read()

        try:
            diffset.finalize_commit_series(
                cumulative_diff=diff_file_contents,
                validation_info=validation_info,
                parent_diff=parent_diff_file_contents,
                request=request,
                save=False)
        except ValidationError as e:
            if e.code == 'invalid':
                return INVALID_ATTRIBUTE, {
                    'reason': e.message,
                }
            elif e.code == 'validation_info':
                return INVALID_FORM_DATA, {
                    'fields': {
                        'validation_info': [e.message],
                    },
                }


draft_diff_resource = DraftDiffResource()



"""Extension hooks for Review Board."""

from djblets.extensions.hooks import (DataGridColumnsHook,
                                      ExtensionHook,
                                      SignalHook,
                                      TemplateHook,
                                      URLHook)
from djblets.privacy.consent.hooks import ConsentRequirementHook

from reviewboard.extensions.hooks.account_page import (AccountPageFormsHook,
                                                       AccountPagesHook)
from reviewboard.extensions.hooks.actions import (
    ActionHook,
    DiffViewerActionHook,
    HeaderActionHook,
    HeaderDropdownActionHook,
    HideActionHook,
    ReviewRequestActionHook,
    ReviewRequestDropdownActionHook)
from reviewboard.extensions.hooks.admin_widget import AdminWidgetHook
from reviewboard.extensions.hooks.api_extra_data_access import \
    APIExtraDataAccessHook
from reviewboard.extensions.hooks.auth_backend import AuthBackendHook
from reviewboard.extensions.hooks.avatar_service import AvatarServiceHook
from reviewboard.extensions.hooks.comment_detail_display import \
    CommentDetailDisplayHook
from reviewboard.extensions.hooks.conditions import \
    ReviewRequestConditionChoicesHook
from reviewboard.extensions.hooks.dashboard import (DashboardColumnsHook,
                                                    DashboardSidebarItemsHook,
                                                    DataGridSidebarItemsHook,
                                                    UserPageSidebarItemsHook)
from reviewboard.extensions.hooks.email import (
    EmailHook,
    ReviewReplyPublishedEmailHook,
    ReviewPublishedEmailHook,
    ReviewRequestClosedEmailHook,
    ReviewRequestPublishedEmailHook)
from reviewboard.extensions.hooks.fields import (ReviewRequestFieldSetsHook,
                                                 ReviewRequestFieldsHook)
from reviewboard.extensions.hooks.file_attachment_thumbnail import \
    FileAttachmentThumbnailHook
from reviewboard.extensions.hooks.filediff_acl import FileDiffACLHook
from reviewboard.extensions.hooks.hosting_service import HostingServiceHook
from reviewboard.extensions.hooks.integration import IntegrationHook
from reviewboard.extensions.hooks.license_provider import LicenseProviderHook
from reviewboard.extensions.hooks.navigation_bar import NavigationBarHook
from reviewboard.extensions.hooks.review_request_approval import \
    ReviewRequestApprovalHook
from reviewboard.extensions.hooks.review_ui import ReviewUIHook
from reviewboard.extensions.hooks.scmtool import SCMToolHook
from reviewboard.extensions.hooks.user_infobox import UserInfoboxHook
from reviewboard.extensions.hooks.users import (ServiceAccountHook,
                                                UserDetailsProviderHook)
from reviewboard.extensions.hooks.webapi_capabilities import \
    WebAPICapabilitiesHook


__all__ = [
    'APIExtraDataAccessHook',
    'AccountPageFormsHook',
    'AccountPagesHook',
    'ActionHook',
    'AdminWidgetHook',
    'AuthBackendHook',
    'AvatarServiceHook',
    'CommentDetailDisplayHook',
    'ConsentRequirementHook',
    'DashboardColumnsHook',
    'DashboardSidebarItemsHook',
    'DataGridColumnsHook',
    'DataGridSidebarItemsHook',
    'DiffViewerActionHook',
    'EmailHook',
    'ExtensionHook',
    'FileAttachmentThumbnailHook',
    'FileDiffACLHook',
    'HeaderActionHook',
    'HeaderDropdownActionHook',
    'HideActionHook',
    'HostingServiceHook',
    'IntegrationHook',
    'LicenseProviderHook',
    'NavigationBarHook',
    'ReviewPublishedEmailHook',
    'ReviewReplyPublishedEmailHook',
    'ReviewRequestActionHook',
    'ReviewRequestApprovalHook',
    'ReviewRequestClosedEmailHook',
    'ReviewRequestConditionChoicesHook',
    'ReviewRequestDropdownActionHook',
    'ReviewRequestFieldSetsHook',
    'ReviewRequestFieldsHook',
    'ReviewRequestPublishedEmailHook',
    'ReviewUIHook',
    'SCMToolHook',
    'ServiceAccountHook',
    'SignalHook',
    'TemplateHook',
    'URLHook',
    'UserDetailsProviderHook',
    'UserInfoboxHook',
    'UserPageSidebarItemsHook',
    'WebAPICapabilitiesHook',
]



"""Management command to manage the search index."""

from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.utils.translation import gettext as _


class Command(BaseCommand):
    """Management command to manage the search index."""

    help = _('Creates a search index of review requests.')
    requires_model_validation = True

    def add_arguments(self, parser):
        """Add arguments to the command.

        Args:
            parser (argparse.ArgumentParser):
                The argument parser for the command.
        """
        parser.add_argument(
            '--full',
            action='store_true',
            dest='rebuild',
            default=False,
            help='Rebuild the database index')

    def handle(self, **options):
        """Handle the command.

        Args:
            **options (dict):
                Options parsed on the command line.
        """
        # Call the appropriate Haystack command to refresh the search index.
        if options['rebuild']:
            call_command('rebuild_index', interactive=False)
        else:
            call_command('update_index')



"""A hook for adding new details to the display of comments."""

from __future__ import annotations

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint


class CommentDetailDisplayHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """This hook allows adding details to the display of comments.

    The hook can provide additional details to display for a comment in a
    review and e-mails.
    """

    def render_review_comment_detail(self, comment):
        """Render additional HTML for a comment on the page.

        Subclasses must implement this to provide HTML for use on the
        review request page or review dialog.

        The result is assumed to be HTML-safe. It's important that subclasses
        escape any data as needed.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to render HTML for,

        Returns:
            django.utils.safestring.SafeText:
            The resulting HTML for the comment. This can be an empty string.
        """
        raise NotImplementedError

    def render_email_comment_detail(self, comment, is_html):
        """Render additional text or HTML for a comment in an e-mail.

        Subclasses must implement this to provide text or HTML (depending on
        the ``is_html`` flag) for use in an e-mail.

        If rendering HTML, the result is assumed to be HTML-safe. It's
        important that subclasses escape any data as needed.

        Args:
            comment (reviewboard.reviews.models.base_comment.BaseComment):
                The comment to render HTML for,

            is_html (bool):
                Whether this must return HTML content.

        Returns:
            django.utils.safestring.SafeText:
            The resulting HTML for the comment. This can be an empty string.
        """
        raise NotImplementedError



import base64
import sys
import time
from datetime import datetime
from hashlib import sha1

from django.conf import settings
from django.contrib.auth.models import User
from django.utils.encoding import force_bytes
from djblets.siteconfig.models import SiteConfiguration

from reviewboard import get_package_version


def get_install_key():
    """Return the installation key for this server."""
    return sha1(force_bytes(settings.SECRET_KEY)).hexdigest()


def _norm_siteconfig_value(siteconfig, key):
    """Normalize site configuration values to strip extra quotation marks."""
    value = siteconfig.get(key)

    # To work around rb-site requiring values in previous releases for these,
    # people would try empty quotes. We want to convert those to actual
    # empty strings.
    if value in ('""', "''", None):
        value = ''

    return value


def serialize_support_data(request=None, force_is_admin=False):
    """Serialize support data into a base64-encoded string."""
    siteconfig = SiteConfiguration.objects.get_current()

    is_admin = (force_is_admin or
                (request is not None and request.user.is_staff))

    return base64.b64encode('\t'.join([
        get_install_key(),
        '%d' % is_admin,
        siteconfig.site.domain,
        _norm_siteconfig_value(siteconfig, 'site_admin_name'),
        _norm_siteconfig_value(siteconfig, 'site_admin_email'),
        get_package_version(),
        '%d' % User.objects.filter(is_active=True).count(),
        '%d' % int(time.mktime(datetime.now().timetuple())),
        _norm_siteconfig_value(siteconfig, 'company'),
        '%s.%s.%s' % sys.version_info[:3],
    ]).encode('utf-8')).decode('utf-8')


def get_default_support_url(request=None, force_is_admin=False):
    """Return the URL for the default Review Board support page."""
    siteconfig = SiteConfiguration.objects.get_current()

    if siteconfig.get('send_support_usage_stats'):
        support_data = serialize_support_data(request, force_is_admin)
    else:
        support_data = ''

    return settings.DEFAULT_SUPPORT_URL % {
        'support_data': support_data,
    }


def get_register_support_url(request=None, force_is_admin=False):
    """Return the URL for registering the Review Board support page."""
    siteconfig = SiteConfiguration.objects.get_current()

    if siteconfig.get('send_support_usage_stats'):
        support_data = serialize_support_data(request, force_is_admin)
    else:
        support_data = ''

    return settings.REGISTER_SUPPORT_URL % {
        'support_data': support_data,
    }


def get_support_url(request):
    """Return the URL for the configured support page."""
    siteconfig = SiteConfiguration.objects.get_current()

    return (siteconfig.get('support_url') or
            get_default_support_url(request))


def get_kb_url(article_id):
    """Return the URL to a knowledge base article on the support tracker.

    Version Added:
        3.0.18

    Args:
        article_id (int):
            The ID of the article.

    Returns:
        unicode:
        The URL to the article.
    """
    return (
        'https://support.beanbaginc.com/support/solutions/articles/%s'
        % article_id
    )



"""Unit tests for reviewboard.accounts.views.ClientLoginConfirmationView.

Version Added:
    5.0.5
"""

from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class ClientLoginConfirmationViewTests(TestCase):
    """Unit tests for reviewboard.accounts.views.ClientLoginConfirmationView.

    Version Added:
        5.0.5
    """

    fixtures = ['test_users']

    #: The URL to the login page.
    login_url = local_site_reverse('login')

    #: The URL to the logout page.
    logout_url = local_site_reverse('logout')

    def test_get(self) -> None:
        """Testing ClientLoginConfirmationView GET"""
        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login-confirm'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                })

        context = rsp.context

        self.assertEqual(context['client_name'], 'TestClient')
        self.assertEqual(context['client_url'], 'http://localhost:1234/test/')
        self.assertEqual(
            context['client_login_url'],
            '/account/client-login/?client-name=TestClient'
            '&client-url=http://localhost:1234/test/')
        self.assertEqual(
            context['logout_url'],
            (f'{self.logout_url}?next={self.login_url}'
             '%3Fclient-name%3DTestClient'
             '%26client-url%3Dhttp%3A//localhost%3A1234/test/'))
        self.assertEqual(context['username'], 'doc')

    def test_get_with_redirect(self) -> None:
        """Testing ClientLoginConfirmationView GET with a redirect URL"""
        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login-confirm'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                    'next': 'http://localhost:1234/page?foo=1',
                })

        context = rsp.context

        self.assertEqual(context['client_name'], 'TestClient')
        self.assertEqual(context['client_url'], 'http://localhost:1234/test/')
        self.assertEqual(
            context['client_login_url'],
            '/account/client-login/?client-name=TestClient'
            '&client-url=http://localhost:1234/test/'
            '&next=http%3A//localhost%3A1234/page%3Ffoo%3D1')

        # The client redirect part of the URL is encoded twice
        # in order to preserve any of its query parameters.
        self.assertEqual(
            context['logout_url'],
            (f'{self.logout_url}?next={self.login_url}'
             '%3Fclient-name%3DTestClient'
             '%26client-url%3Dhttp%3A//localhost%3A1234/test/'
             '%26next%3Dhttp%253A//localhost%253A1234/page%253Ffoo%253D1'))
        self.assertEqual(context['username'], 'doc')

    def test_get_with_unsafe_redirect(self) -> None:
        """Testing ClientLoginConfirmationView GET with an unsafe
        redirect URL
        """
        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login-confirm'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                    'next': 'http://unsafe-site/page?foo=1',
                })

        context = rsp.context

        self.assertEqual(context['client_name'], 'TestClient')
        self.assertEqual(context['client_url'], 'http://localhost:1234/test/')
        self.assertEqual(
            context['client_login_url'],
            '/account/client-login/?client-name=TestClient'
            '&client-url=http://localhost:1234/test/')
        self.assertEqual(
            context['logout_url'],
            (f'{self.logout_url}?next={self.login_url}'
             '%3Fclient-name%3DTestClient'
             '%26client-url%3Dhttp%3A//localhost%3A1234/test/'))
        self.assertEqual(context['username'], 'doc')

    def test_get_unauthenticated(self) -> None:
        """Testing ClientLoginConfirmationView GET redirects to the
        login page when a user is not logged in
        """
        settings = {
            'client_web_login': True,
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login-confirm'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                })

        self.assertEqual(rsp.status_code, 302)

    def test_get_client_web_login_false(self) -> None:
        """Testing ClientLoginConfirmationView GET with the client
        web login flow disabled
        """
        settings = {
            'client_web_login': False,
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.get(
                local_site_reverse('client-login-confirm'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                })

        self.assertEqual(rsp.status_code, 404)

    def test_post(self) -> None:
        """Testing ClientLoginConfirmationView POST"""
        self.client.login(username='doc', password='doc')
        settings = {
            'client_web_login': True,
        }

        with self.siteconfig_settings(settings):
            rsp = self.client.post(
                local_site_reverse('client-login-confirm'),
                {
                    'client-name': 'TestClient',
                    'client-url': 'http://localhost:1234/test/',
                })

        self.assertEqual(rsp.status_code, 405)



"""The app definition for reviewboard.attachments.

Version Added:
    6.0
"""

from __future__ import annotations

from django.apps import AppConfig


class AttachmentsAppConfig(AppConfig):
    """App configuration for reviewboard.attachments.

    Version Added:
        6.0
    """

    name = 'reviewboard.attachments'

    def ready(self) -> None:
        """Configure the app once it's ready.

        This will connect signal handlers for the app.
        """
        from reviewboard.attachments.signal_handlers import \
            connect_signal_handlers

        connect_signal_handlers()



"""Unit tests for the DiffResource."""

from __future__ import annotations

import kgb
from django.core.files.uploadedfile import SimpleUploadedFile
from djblets.features.testing import override_feature_check
from djblets.webapi.errors import (INVALID_ATTRIBUTE, INVALID_FORM_DATA,
                                   PERMISSION_DENIED)
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.diffviewer.models import DiffSet
from reviewboard.reviews.models import DefaultReviewer
from reviewboard.reviews.signals import review_request_diffset_uploaded
from reviewboard.webapi.errors import DIFF_TOO_BIG
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (diff_item_mimetype,
                                                diff_list_mimetype)
from reviewboard.webapi.tests.mixins import (BasicTestsMetaclass,
                                             ReviewRequestChildItemMixin,
                                             ReviewRequestChildListMixin)
from reviewboard.webapi.tests.mixins_extra_data import (ExtraDataItemMixin,
                                                        ExtraDataListMixin)
from reviewboard.webapi.tests.urls import (get_diff_item_url,
                                           get_diff_list_url)


class ResourceListTests(kgb.SpyAgency, ExtraDataListMixin,
                        ReviewRequestChildListMixin, BaseWebAPITestCase,
                        metaclass=BasicTestsMetaclass):
    """Testing the DiffResource list APIs."""
    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-requests/<id>/diffs/'
    resource = resources.diff

    def setup_review_request_child_test(self, review_request):
        return get_diff_list_url(review_request), diff_list_mimetype

    def setup_http_not_allowed_item_test(self, user):
        review_request = self.create_review_request(
            create_repository=True,
            submitter=user,
            publish=True)

        return get_diff_list_url(review_request)

    def compare_item(self, item_rsp, diffset):
        self.assertEqual(item_rsp['id'], diffset.pk)
        self.assertEqual(item_rsp['name'], diffset.name)
        self.assertEqual(item_rsp['revision'], diffset.revision)
        self.assertEqual(item_rsp['basedir'], diffset.basedir)
        self.assertEqual(item_rsp['base_commit_id'], diffset.base_commit_id)
        self.assertEqual(item_rsp['extra_data'], diffset.extra_data)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user,
            publish=True)

        if populate_items:
            items = [self.create_diffset(review_request)]
        else:
            items = []

        return (get_diff_list_url(review_request, local_site_name),
                diff_list_mimetype,
                items)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        repository = self.create_repository(tool_name='Test')
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            repository=repository,
            submitter=user)

        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        if post_valid_data:
            post_data = {
                'path': diff,
                'basedir': '/trunk',
                'base_commit_id': '1234',
            }
        else:
            post_data = {}

        return (get_diff_list_url(review_request, local_site_name),
                diff_item_mimetype,
                post_data,
                [review_request])

    def check_post_result(self, user, rsp, review_request):
        self.assertIn('diff', rsp)
        item_rsp = rsp['diff']

        draft = review_request.get_draft()
        self.assertIsNotNone(draft)

        diffset = DiffSet.objects.get(pk=item_rsp['id'])
        self.assertEqual(diffset, draft.diffset)
        self.compare_item(item_rsp, diffset)

    def test_post_with_missing_data(self):
        """Testing the POST review-requests/<id>/diffs/ API
        with Invalid Form Data
        """
        repository = self.create_repository(tool_name='Test')
        review_request = self.create_review_request(
            repository=repository,
            submitter=self.user)

        rsp = self.api_post(get_diff_list_url(review_request),
                            expected_status=400)
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertIn('path', rsp['fields'])

        # Now test with a valid path and an invalid basedir.
        # This is necessary because basedir is "optional" as defined by
        # the resource, but may be required by the form that processes the
        # diff.
        review_request = self.create_review_request(
            repository=repository,
            submitter=self.user)

        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        rsp = self.api_post(
            get_diff_list_url(review_request),
            {'path': diff},
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertIn('basedir', rsp['fields'])

    def test_post_too_big(self):
        """Testing the POST review-requests/<id>/diffs/ API
        with diff exceeding max size
        """
        repository = self.create_repository()
        review_request = self.create_review_request(
            repository=repository,
            submitter=self.user)

        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        with self.siteconfig_settings({'diffviewer_max_diff_size': 2},
                                      reload_settings=False):
            rsp = self.api_post(
                get_diff_list_url(review_request),
                {
                    'path': diff,
                    'basedir': "/trunk",
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DIFF_TOO_BIG.code)
        self.assertIn('reason', rsp)
        self.assertIn('max_size', rsp)
        self.assertEqual(rsp['max_size'], 2)

    def test_post_not_owner(self):
        """Testing the POST review-requests/<id>/diffs/ API
        without owner
        """
        repository = self.create_repository(tool_name='Test')
        review_request = self.create_review_request(repository=repository)

        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        rsp = self.api_post(
            get_diff_list_url(review_request),
            {
                'path': diff,
                'basedir': '/trunk',
            },
            expected_status=403)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    def test_post_no_repository(self):
        """Testing the POST review-requests/<id>/diffs API
        with a ReviewRequest that has no repository
        """
        review_request = self.create_review_request(submitter=self.user)

        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        rsp = self.api_post(
            get_diff_list_url(review_request),
            {
                'path': diff,
                'basedir': '/trunk',
            },
            expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_ATTRIBUTE.code)

    @webapi_test_template
    def test_post_with_history(self):
        """Testing the POST <URL> API with a diff and a review request created
        with history support
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True,
                                                    create_with_history=True)

        diff = SimpleUploadedFile('diff',
                                  self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
                                  content_type='text/x-patch')

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            rsp = self.api_post(
                get_diff_list_url(review_request),
                {
                    'path': diff,
                },
                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertEqual(
            rsp['reason'],
            'This review request was created with support for multiple '
            'commits.\n\n'
            'Create an empty diff revision and upload commits to that '
            'instead.')

    @webapi_test_template
    def test_post_empty_with_history(self):
        """Testing the POST <URL> API creates an empty DiffSet for a review
        request created with history support with the DVCS feature enabled
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True,
                                                    create_with_history=True)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            rsp = self.api_post(get_diff_list_url(review_request), {},
                                expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        item_rsp = rsp['diff']

        diff = DiffSet.objects.get(pk=item_rsp['id'])
        self.compare_item(item_rsp, diff)
        self.assertEqual(diff.files.count(), 0)
        self.assertEqual(diff.revision, 1)

    @webapi_test_template
    def test_post_empty_dvcs_disabled(self):
        """Testing the POST <URL> API without a diff with the DVCS feature
        disabled
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True,
                                                    create_with_history=False)

        with override_feature_check(dvcs_feature.feature_id, enabled=False):
            rsp = self.api_post(get_diff_list_url(review_request), {},
                                expected_status=400)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)
        self.assertEqual(rsp['fields'], {
            'path': ['This field is required.'],
        })

    @webapi_test_template
    def test_post_adds_default_reviewers(self):
        """Testing the POST <URL> API adds default reviewers"""
        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True)

        # Create the state needed for the default reviewer.
        group = self.create_review_group(name='group1')

        default_reviewer = DefaultReviewer.objects.create(
            name='default1',
            file_regex='.')
        default_reviewer.groups.add(group)
        default_reviewer.repository.add(review_request.repository)

        # Post the diff.
        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        rsp = self.api_post(
            get_diff_list_url(review_request),
            {'path': diff},
            expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        draft = review_request.get_draft()
        self.assertEqual(list(draft.target_groups.all()), [group])

    @webapi_test_template
    def test_post_adds_default_reviewers_first_time_only(self):
        """Testing the POST <URL> API doesn't add default reviewers a second
        time
        """
        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True)

        # Create the initial diffset. This should prevent a default
        # reviewer from being applied, since we're not publishing the first
        # diff on a review request.
        self.create_diffset(review_request=review_request)

        # Create the state needed for the default reviewer.
        group = self.create_review_group(name='group1')

        default_reviewer = DefaultReviewer.objects.create(
            name='default1',
            file_regex='.')
        default_reviewer.groups.add(group)
        default_reviewer.repository.add(review_request.repository)

        # Post the diff.
        diff = SimpleUploadedFile('diff', self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        rsp = self.api_post(
            get_diff_list_url(review_request),
            {'path': diff},
            expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

        draft = review_request.get_draft()
        self.assertEqual(list(draft.target_groups.all()), [])

    @webapi_test_template
    def test_post_emits_review_request_diffset_uploaded(self) -> None:
        """Testing the POST <URL> API emits a review_request_diffset_uploaded
        signal
        """
        def on_diffset_uploaded(diffset, review_request_draft, **kwargs):
            pass

        review_request_diffset_uploaded.connect(on_diffset_uploaded)
        self.spy_on(on_diffset_uploaded)

        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True)

        diff = SimpleUploadedFile('diff',
                                  self.DEFAULT_GIT_README_DIFF,
                                  content_type='text/x-patch')

        rsp = self.api_post(
            get_diff_list_url(review_request),
            {
                'path': diff
            },
            expected_mimetype=diff_item_mimetype)

        draft = review_request.get_draft()
        diffset = DiffSet.objects.get(pk=rsp['diff']['id'])

        self.assertEqual(rsp['stat'], 'ok')
        self.assertSpyCalledWith(
            on_diffset_uploaded,
            diffset=diffset,
            review_request_draft=draft)

        review_request_diffset_uploaded.disconnect(on_diffset_uploaded)

    @webapi_test_template
    def test_post_with_history_review_request_diffset_uploaded(self) -> None:
        """Testing the POST <URL> API does not emit a
        review_request_diffset_uploaded signal for a review request created
        with history support
        """
        def on_diffset_uploaded(diffset, review_request_draft, **kwargs):
            pass

        review_request_diffset_uploaded.connect(on_diffset_uploaded)
        self.spy_on(on_diffset_uploaded)

        review_request = self.create_review_request(submitter=self.user,
                                                    create_repository=True,
                                                    create_with_history=True)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            rsp = self.api_post(get_diff_list_url(review_request), {},
                                expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertSpyNotCalled(on_diffset_uploaded)

        review_request_diffset_uploaded.disconnect(on_diffset_uploaded)


class ResourceItemTests(ExtraDataItemMixin, ReviewRequestChildItemMixin,
                        BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the DiffResource item APIs."""
    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'review-requests/<id>/diffs/<revision>/'
    resource = resources.diff

    def setup_review_request_child_test(self, review_request):
        if not review_request.repository:
            review_request.repository = self.create_repository()
            review_request.save()

        diffset = self.create_diffset(review_request)

        return (get_diff_item_url(review_request, diffset.revision),
                diff_item_mimetype)

    def setup_http_not_allowed_item_test(self, user):
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        return get_diff_item_url(review_request, 1)

    def setup_http_not_allowed_list_test(self, user):
        review_request = self.create_review_request(
            create_repository=True,
            submitter=user)
        diffset = self.create_diffset(review_request)

        return get_diff_item_url(review_request, diffset.revision)

    def compare_item(self, item_rsp, diffset):
        self.assertEqual(item_rsp['id'], diffset.pk)
        self.assertEqual(item_rsp['name'], diffset.name)
        self.assertEqual(item_rsp['revision'], diffset.revision)
        self.assertEqual(item_rsp['basedir'], diffset.basedir)
        self.assertEqual(item_rsp['base_commit_id'], diffset.base_commit_id)
        self.assertEqual(item_rsp['extra_data'], diffset.extra_data)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user)
        diffset = self.create_diffset(review_request)

        return (get_diff_item_url(review_request, diffset.revision,
                                  local_site_name),
                diff_item_mimetype,
                diffset)

    def test_get_not_modified(self):
        """Testing the GET review-requests/<id>/diffs/<revision>/ API
        with Not Modified response
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)

        self._testHttpCaching(
            get_diff_item_url(review_request, diffset.revision),
            check_etags=True)

    @webapi_test_template
    def test_get_with_patch_and_commit_history(self):
        """Testing the GET <API> API with Accept: x-patch and commit history
        contains only cumulative diff
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request=review_request)

        self.create_diffcommit(
            diffset=diffset,
            commit_id='r1',
            parent_id='r0',
            diff_contents=(
                b'diff --git a/ABC b/ABC\n'
                b'index 94bdd3e..197009f 100644\n'
                b'--- ABC\n'
                b'+++ ABC\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-line!\n'
                b'+line..\n'
            ))
        self.create_diffcommit(
            diffset=diffset,
            commit_id='r2',
            parent_id='r1',
            diff_contents=(
                b'diff --git a/README b/README\n'
                b'index 94bdd3e..197009f 100644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hello, world!\n'
                b'+Hi, world!\n'
            ))
        self.create_diffcommit(
            diffset=diffset,
            commit_id='r4',
            parent_id='r3',
            diff_contents=(
                b'diff --git a/README b/README\n'
                b'index 197009f..87abad9 100644\n'
                b'--- README\n'
                b'+++ README\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-Hi, world!\n'
                b'+Yo, world.\n'
            ))

        cumulative_diff = (
            b'diff --git a/ABC b/ABC\n'
            b'index 94bdd3e..197009f 100644\n'
            b'--- ABC\n'
            b'+++ ABC\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-line!\n'
            b'+line..\n'
            b'diff --git a/README b/README\n'
            b'index 94bdd3e..87abad9 100644\n'
            b'--- README\n'
            b'+++ README\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-Hello, world!\n'
            b'+Yo, world.\n'
        )

        diffset.finalize_commit_series(
            cumulative_diff=cumulative_diff,
            validation_info=None,
            validate=False,
            save=True)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            rsp = self.api_get(get_diff_item_url(review_request,
                                                 diffset.revision),
                               HTTP_ACCEPT='text/x-patch',
                               expected_json=False,
                               expected_mimetype='text/x-patch')

        self.assertEqual(rsp, cumulative_diff)

    @webapi_test_template
    def test_get_links_fields_dvcs_enabled(self):
        """Testing the GET <URL> API does includes DVCS-specific fields and
        links when the DVCS feature is enabled
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)

        with override_feature_check(dvcs_feature.feature_id, enabled=True):
            rsp = self.api_get(get_diff_item_url(review_request,
                                                 diffset.revision),
                               expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff', rsp)

        item_rsp = rsp['diff']
        self.assertIn('links', item_rsp)
        self.assertIn('commits', item_rsp['links'])

        self.assertIn('commit_count', item_rsp)

    @webapi_test_template
    def test_get_links_fields_dvcs_disabled(self):
        """Testing the GET <URL> API does not includes DVCS-specific fields and
        links when the DVCS feature is enabled
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)

        with override_feature_check(dvcs_feature.feature_id, enabled=False):
            rsp = self.api_get(get_diff_item_url(review_request,
                                                 diffset.revision),
                               expected_mimetype=diff_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff', rsp)

        item_rsp = rsp['diff']
        self.assertIn('links', item_rsp)
        self.assertNotIn('commits', item_rsp['links'])
        self.assertNotIn('commit_count', item_rsp)

    @webapi_test_template
    def test_get_patch(self):
        """Testing the GET <URL> API with Accept: text/x-patch"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        self.create_filediff(diffset)

        diff, response = self.api_get_with_response(
            get_diff_item_url(review_request, diffset.revision),
            expected_mimetype='text/x-patch',
            expected_json=False,
            HTTP_ACCEPT='text/x-patch')

        self.assertEqual(
            diff,
            b'--- README\trevision 123\n'
            b'+++ README\trevision 123\n'
            b'@@ -1 +1 @@\n'
            b'-Hello, world!\n'
            b'+Hello, everybody!\n')
        self.assertEqual(response['Content-Disposition'],
                         'inline; filename=diffset')

    @webapi_test_template
    def test_get_patch_and_inaccessible(self) -> None:
        """Testing the GET <URL> API with Accept: text/x-patch and
        inaccessible diff
        """
        repository = self.create_repository(public=False)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        self.create_filediff(diffset)

        rsp = self.api_get(
            get_diff_item_url(review_request, diffset.revision),
            HTTP_ACCEPT='text/x-patch',
            expected_status=403)

        self.assertEqual(
            rsp,
            {
                'err': {
                    'code': PERMISSION_DENIED.code,
                    'msg': PERMISSION_DENIED.msg,
                    'type': 'resource-permission-denied',
                },
                'stat': 'fail',
            })

    @webapi_test_template
    def test_get_patch_with_bugs_closed(self):
        """Testing the GET <URL> API with Accept: text/x-patch and bugs_closed
        field
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True,
                                                    bugs_closed='123,456')
        diffset = self.create_diffset(review_request, name='diff')
        self.create_filediff(diffset)

        diff, response = self.api_get_with_response(
            get_diff_item_url(review_request, diffset.revision),
            expected_mimetype='text/x-patch',
            expected_json=False,
            HTTP_ACCEPT='text/x-patch')

        self.assertEqual(
            diff,
            b'--- README\trevision 123\n'
            b'+++ README\trevision 123\n'
            b'@@ -1 +1 @@\n'
            b'-Hello, world!\n'
            b'+Hello, everybody!\n')
        self.assertEqual(response['Content-Disposition'],
                         'inline; filename=bug123_456.patch')

    @webapi_test_template
    def test_get_patch_with_bugs_closed_with_newline(self):
        """Testing the GET <URL> API with Accept: text/x-patch and bugs_closed
        field with a newline in it.
        """
        # We had a report that someone was copy/pasting bug numbers into the
        # bugs field, and managed to paste something that included a newline
        # character. This then failed when attempting to use that bug number in
        # the Content-Disposition header.
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True,
                                                    bugs_closed='123,\n456')
        diffset = self.create_diffset(review_request, name='diff')
        self.create_filediff(diffset)

        diff, response = self.api_get_with_response(
            get_diff_item_url(review_request, diffset.revision),
            expected_mimetype='text/x-patch',
            expected_json=False,
            HTTP_ACCEPT='text/x-patch')

        self.assertEqual(
            diff,
            b'--- README\trevision 123\n'
            b'+++ README\trevision 123\n'
            b'@@ -1 +1 @@\n'
            b'-Hello, world!\n'
            b'+Hello, everybody!\n')
        self.assertEqual(response['Content-Disposition'],
                         'inline; filename=bug123_456.patch')

    #
    # HTTP PUT tests
    #

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user)
        diffset = self.create_diffset(review_request)

        return (get_diff_item_url(review_request, diffset.revision,
                                  local_site_name),
                diff_item_mimetype,
                {},
                diffset,
                [])

    def check_put_result(self, user, item_rsp, diffset):
        diffset = DiffSet.objects.get(pk=diffset.pk)
        self.compare_item(item_rsp, diffset)
